"""Utility script for training a PhageClassifier."""

import argparse
import pathlib

import pandas as pd

from depht.classes.prophage_classifier import ProphageClassifier
from depht_train.classes.kfold import KFold


def parse_args():
    """Parse commandline arguments."""
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument("phage_csv", type=pathlib.Path,
                   help="path to CSV file containing phage data")
    p.add_argument("bacteria_csv", type=pathlib.Path,
                   help="path to CSV file containing bacterial data")
    p.add_argument("classifier_path", type=pathlib.Path,
                   help="path where classifier should be stored")

    p.add_argument("-k", "--k-fold", type=int, default=5,
                   help=f"perform [k]-fold cross validation [default: {5}]")
    p.add_argument("-w", "--optimize-weights", action="store_true",
                   help="optimize feature weights for the trained model")
    p.add_argument("-t", "--optimize-threshold", action="store_true",
                   help="optimize threshold for the model to predict 0/1")

    return p.parse_args()


def even_weights(n, k=100):
    """Returns the feature weights that weigh `n` features evenly.

    :param n: the number of features to weigh evenly
    :type n: int
    :param k: value the weights should sum to
    :type k: int
    """
    return [round(float(k) / n, 2)] * n


def multiset(n, k=100, k_step=5):
    """Returns the multiset of `n` integers that sum to `k` given a
    step size of `k_step`.

    Useful for generating feature weight parameters to test, given a
    modest `n`, `k`=100 and `k_step` an even divisor of 100 (e.g. 1,
    2, 5, 10, 20, 25, 50).

    :param n: the number of features being trained against
    :type n: int
    :param k: sum of feature weights must equal this number
    :type k: int
    :param k_step: granularity for changing feature weights
    :type k_step: int
    """
    # Neither n, nor k can be negative
    if n < 0 or k < 0:
        raise ValueError(f"map_feature_weights requires (k, n) >= (0, 0): "
                         f"({n}, {k}) invalid")

    if not n:
        return []

    if not k:
        return [[0] * n]

    if n == 1:
        return [[k]]

    return [[0] + v for v in multiset(n - 1, k)] + \
           [[v[0] + k_step] + v[1:] for v in multiset(n, k - k_step)]


def mcc(true_pos, false_neg, true_neg, false_pos):
    """Calculate the Matthews correlation coefficient.

    NOTE: this can be thought of as a balanced accuracy metric
    that isn't skewed by differences in class size.

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: mcc
    """
    numer = (true_pos * true_neg) + (false_pos * false_neg)
    denom = ((true_pos + false_pos) * (true_pos + false_neg) *
             (true_neg + false_pos) * (true_neg + false_neg)) ** 0.5

    return float(numer)/denom


def score(real_labels, predict_labels):
    """
    Helper function to score class predictions and then return
    the mcc and f1_scores.
    """
    tps, tns, fps, fns = 0, 0, 0, 0

    for real_value, predict_value in zip(real_labels, predict_labels):
        if real_value and predict_value:
            tps += 1
        elif real_value and not predict_value:
            fns += 1
        elif predict_value and not real_value:
            fps += 1
        else:
            tns += 1

    print(f"tps: {tps}\tfns: {fns}\ttns: {tns}\tfps: {fps}")

    return mcc(tps, fns, tns, fps)


def train_classifier(df, k=5, verbose=False):
    """Train a ProphageClassifier.

    :param df: concatenated phage genomes dataframe
    :type df: pandas.DataFrame
    :param k: fold validation to perform
    :type k: int
    :param verbose:
    :type verbose: bool
    :return: classifier, feature_weights, threshold
    """
    phg_df, bct_df = df[df["class"] == 1], df[df["class"] == 0]

    clf = ProphageClassifier()

    # Seed RNG for reproducibility - use the "Answer to the Ultimate
    # Question of Life, the Universe, and Everything." (Douglas Adams)
    kf = KFold(n_splits=k, shuffle=True, random_state=42)

    phg_splits, bct_splits = kf.split(len(phg_df)), kf.split(len(bct_df))

    evaluations = dict()
    for phg_split, bct_split in zip(phg_splits, bct_splits):
        p_train, p_test = phg_split
        b_train, b_test = bct_split

        train_feats = pd.concat((phg_df.iloc[list(p_train), :-1],
                                 bct_df.iloc[list(b_train), :-1]), axis=0)
        train_labels = pd.concat((phg_df.iloc[list(p_train), -1],
                                  bct_df.iloc[list(b_train), -1]), axis=0)
        test_feats = pd.concat((phg_df.iloc[list(p_test), :-1],
                                bct_df.iloc[list(b_test), :-1]), axis=0)
        test_labels = pd.concat((phg_df.iloc[list(p_test), -1],
                                 bct_df.iloc[list(b_test), -1]), axis=0)

        clf.fit(train_feats, train_labels)

        # Optimize feature weights
        for fw in multiset(n=len(clf), k=100, k_step=5):
            fw = [x / 100 for x in fw]
            predictions = clf.predict(test_feats, feature_weights=fw, alpha=0.5)
            mcc_score = score(test_labels, predictions)

            key = tuple(fw)
            if key in evaluations:
                evaluations[key].append(mcc_score)
            else:
                evaluations[key] = [mcc_score]

            if verbose:
                print(f"{key}: {mcc_score}")

    ranked_params = list()
    for key, values in evaluations.items():
        value = sum(values) / k
        ranked_params.append((value, key))
    ranked_params = sorted(ranked_params)

    best_params = ranked_params[-1]
    opt_mcc = best_params[0]
    opt_weights = best_params[1]

    if verbose:
        print(f"The best weights ({opt_weights}) get average MCC = "
              f"{round(opt_mcc, 5)}")

    all_feats = pd.concat((phg_df.iloc[:, :-1],
                           bct_df.iloc[:, :-1]), axis=0)
    all_labels = pd.concat((phg_df.iloc[:, -1],
                            bct_df.iloc[:, -1]), axis=0)

    figs = clf.fit(all_feats, all_labels, plot=True)

    return clf, figs



"""Curate phage gene phamilies, looking for conserved functional
annotations.
"""
import argparse
import pathlib
import sys

from Bio import SeqIO, AlignIO

from depht.data import GLOBAL_VARIABLES
from depht.functions.multiprocess import CPUS, parallelize
from depht_train.data import PARAMETERS
from depht_train.functions import clustalo, fileio

MIN_HMM_COUNT = PARAMETERS["phage_homologs"]["min_hmm_count"]
AC_THRESHOLD = PARAMETERS["phage_homologs"]["annotation_consensus_threshold"]


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def parse_args(unparsed_args):
    """Parse commandline arguments."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("input_dir", type=pathlib.Path)
    parser.add_argument("output_dir", type=pathlib.Path)
    parser.add_argument("index_file", type=pathlib.Path)

    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-n", "--name", type=str,
                        default=GLOBAL_VARIABLES["phage_sequences"]["name"])
    parser.add_argument("-np", "--cpus", type=int, default=CPUS)
    parser.add_argument("-c", "--config", type=pathlib.Path, default=None)

    parser.add_argument("-ct", "--consensus_threshold", type=float,
                        default=AC_THRESHOLD)
    parser.add_argument("-all", "--accept_all", action="store_true")
    parser.add_argument("-mc", "--min_hmm_count", type=int)

    args = parser.parse_args(unparsed_args)
    return args


def curate_gene_clusters(fasta_dir, index_file, output_dir,
                         accepted_functions, ignored_functions,
                         accept_all=False, verbose=False, cores=1,
                         min_hmm_count=MIN_HMM_COUNT,
                         consensus_threshold=AC_THRESHOLD):
    """Annotate and curate gene clusters, discarding those gene clusters
    with undesired or unclear consensus functions

    :param accept_all:
    :param fasta_dir: Path to dir containing protein cluster fasta-alignments
    :type fasta_dir: pathlib.Path
    :param index_file: Path to gene index and relevant metadata file
    :type index_file: pathlib.Path
    :param accepted_functions: List of function annotations to accept
    :type accepted_functions: list[str]
    :param ignored_functions: List of function annotations to ignore
    :type ignored_functions: list[str]
    :param output_dir: Path to directory to dump named/aligned sequence files
    :type output_dir: pathlib.Path
    :param cores: Number of processes to utilize
    :type cores: int
    :param verbose: Toggle pipeline print statements
    :type verbose: bool
    :param min_hmm_count: Minimum size threshold for protein clusters
    :type min_hmm_count: int
    :param consensus_threshold: Fraction required for the consensus product.
    :type consensus_threshold: float
    """
    # Annotate gene clusters based on consensus product annotation
    cluster_functions = annotate_gene_clusters(
                                    fasta_dir, index_file,
                                    consensus_threshold=consensus_threshold)

    # Create a lookup map of clusters by function
    function_to_cluster_map = map_function_to_cluster(cluster_functions)

    # Isolate clusters whose consensus product annotation
    # meets the desired annotations from the config file
    curated_clusters = []
    for cluster_function, clusters in function_to_cluster_map.items():
        ignore = False
        for ignored_function in ignored_functions:
            if ignored_function.lower() in cluster_function.lower():
                ignore = True
                break

        if ignore:
            continue

        if not accept_all:
            for accepted_function in accepted_functions:
                if accepted_function.lower() in cluster_function.lower():
                    curated_clusters += clusters
                    break

            continue

        curated_clusters += clusters

    output_dir.mkdir(exist_ok=True, parents=True)
    # Prepare jobs for protein sequence alignment
    work_items = []
    for cluster in curated_clusters:
        cluster_path = fasta_dir.joinpath(cluster)
        outpath = output_dir.joinpath(cluster)

        records = [record for record in SeqIO.parse(cluster_path, "fasta")]

        if len(records) < min_hmm_count:
            continue

        work_items.append((records, outpath, cluster_functions[cluster]))

    # Run jobs
    parallelize(work_items, cores, dump_named_alignment, verbose=verbose)


def dump_named_alignment(records, fasta_path, name):
    """Function to align a multiple sequence file and provide a name
    according to the specifications for hhsuite database building.

    :param records: SeqRecord protein sequence objects to be aligned
    :type records: list
    :param fasta_path: Path where the alignment file will be written
    :type fasta_path: pathlib.Path
    :param name: Name given to the seqeunce alignment
    :type name: str
    """
    with fasta_path.open(mode="w") as filehandle:
        SeqIO.write(records, filehandle, "fasta")

    clustalo.clustalo(fasta_path, fasta_path)

    aln_records = [record for record in AlignIO.read(fasta_path, "fasta")]

    with fasta_path.open(mode="w") as filehandle:
        filehandle.write("".join(["#", name, "\n"]))

        SeqIO.write(aln_records, filehandle, "fasta")


def map_function_to_cluster(cluster_functions):
    """

    :param cluster_functions:
    :return:
    """
    function_to_cluster_map = {}
    for cluster_name, cluster_function in cluster_functions.items():
        mapped_clusters = function_to_cluster_map.get(cluster_function, list())
        mapped_clusters.append(cluster_name)
        function_to_cluster_map[cluster_function] = mapped_clusters

    return function_to_cluster_map


def annotate_gene_clusters(fasta_dir, index_file,
                           consensus_threshold=AC_THRESHOLD):
    """Function to annotate a protein sequence cluster based on
    the product annotations of its members

    :param fasta_dir: Path to dir containing protein cluster fasta-alignments
    :type fasta_dir: pathlib.Path
    :param index_file: Path to gene index and relevant metadata file
    :type index_file: pathlib.Path
    :param consensus_threshold:  Fraction required for the consensus product
    :type consensus_threshold: float
    """
    gene_index = fileio.read_gene_index_file(index_file)

    cluster_functions = dict()
    for input_file in fasta_dir.iterdir():
        records = [record for record in SeqIO.parse(input_file, "fasta")]

        product_counts = {}
        for record in records:
            index = record.id

            gene_data = gene_index.get(index)
            if gene_data is None:
                continue

            product_count = product_counts.get(gene_data["product"], 0)
            product_count += 1
            product_counts[gene_data["product"]] = product_count

        products = [product for product, count in product_counts.items()
                    if (count / len(records)) >= consensus_threshold]
        products.sort(key=lambda x: product_counts[x], reverse=True)

        cluster_product = GLOBAL_VARIABLES["sequences"]["default_product"]
        for product in products:
            if product == GLOBAL_VARIABLES["sequences"]["default_product"]:
                continue

            cluster_product = product
            break

        cluster_functions[input_file.name] = cluster_product

    return cluster_functions


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    accept_list = None
    ignore_list = None
    if args.config is not None:
        if args.config.is_file():
            # Read desired product annotation configuration file
            accept_list, ignore_list = fileio.read_functions_config_file(
                                                        args.config)

    if accept_list is None or ignore_list is None:
        if accept_list is None:
            accept_list = PARAMETERS["phage_homologs"][
                                     "essential_annotations"]["LIKE"]
        if ignore_list is None:
            ignore_list = PARAMETERS["phage_homologs"][
                                     "essential_annotations"]["NOT LIKE"]

    curate_gene_clusters(
        args.input_dir, args.index_file, args.output_dir,
        accept_list, ignore_list, cores=args.cpus, verbose=args.verbose,
        accept_all=args.accept_all, min_hmm_count=args.min_hmm_count,
        consensus_threshold=args.consensus_threshold)


if __name__ == "__main__":
    main()



"""Report comparative statistics, scoring DEPhT and other tools for
their accuracy and precision in identifying and extracting prophages.
"""
import argparse
import csv
import math
import pathlib
import sys

from Bio import SeqIO

# Global variables
CSV_HEADERS = ["strain", "software", "prophage", "ends"]    # headers for csv
GENOME_LENGTHS = {}     # stores lengths of the genomes


# Will convert the file to a biopython object
def open_file(filename, store_length=False):
    """Convert the FASTA formatted file into a biopython object.

    :param filename: name of the file
    :type filename: Path
    :param store_length: True if the lengths need to be stored
    :type store_length: bool
    :return: SeqRecord object generated from the FASTA file
    :rtype: Bio.SeqRecord.SeqRecord
    """
    # iterate through all the contigs
    # this makes sure even extrachromosomal prophages are seen
    contigs = [record for record in SeqIO.parse(filename, "fasta")]

    length = 0

    # length of the genome = sum of lengths of every record
    for record in contigs:
        length += len(record)

    # store lengths of the genomes in the dictionary
    if store_length:
        GENOME_LENGTHS[filename.stem] = length

    return contigs


def find_sub(parent, child_contigs):
    """Locate the prophage region as a subsequence of the bacterial genome.

    :param parent: Parent sequence
    :type parent: Bio.SeqRecord.SeqRecord
    :param other: Subsequence being located in the parent
    :type other: Bio.SeqRecord.SeqRecord
    :return: left end of the subsequence
    :rtype: int
    """
    # keep track of which contig it is in

    for record in child_contigs:
        # Find the beginning of the subsequence
        left_coor = parent.seq.find(record.seq)
        # Checks reverse complement
        if left_coor == -1:
            reverse = record.reverse_complement()
            left_coor = parent.seq.find(reverse.seq)
            # if nothing is returned an error exists
            if left_coor == -1:
                return None
    return left_coor


def stats(manual_data, testing_data):
    """Return True Positive, False Positive, True Negative, False Negative.

    :param manual_coordinates: Manual prophage names w/ Coordinates
    :type manual_coordinates: dict
    :param software_coordinates: Software prophages names w/ Coordinates
    :type software_coordinates: dict
    :return: Dictionary of TP, TF, FP and FN values
    :rtype: dict
    """
    statistics = {}
    statistics["TRUE_POSITIVE"] = 0
    statistics["FALSE_POSITIVE"] = 0
    statistics["FALSE_NEGATIVE"] = 0
    manual_set = set()
    soft_set = set()

    for ends in testing_data.values():  # iterates over the ends
        ends = eval(ends)
        for coor in range(int(ends[0]), int(ends[1])):
            soft_set.add(coor)
    for ends in manual_data.values():  # iterates over the ends for manual
        ends = eval(ends)
        for coor in range(int(ends[0]), int(ends[1])):
            manual_set.add(coor)

    statistics["TRUE_POSITIVE"] += len(soft_set.intersection(manual_set))
    statistics["FALSE_POSITIVE"] += len(soft_set.difference(manual_set))
    statistics["FALSE_NEGATIVE"] += len(manual_set.difference(soft_set))
    return statistics


def collect_stats(filepath, per_strain=False):
    """Compile raw bp values of TP,FP,TN,FN.

    :param filepath: Filepath of the csv file with data entries
    :type filepath: Path
    :return: Dictionary containing TP, TN, FP, FN
    :rtype: dict
    """
    testing_data_title = filepath.stem.split("_")[0]

    true_positive = 0  # the bp of true positives
    true_negative = 0  # the bp of true negatives
    false_positive = 0  # the bp of false positives
    false_negative = 0  # the bp of
    length_manual = 0
    length_test = 0
    statistics = {}

    statistics["TRUE_POSITIVE"] = 0
    statistics["FALSE_POSITIVE"] = 0
    statistics["FALSE_NEGATIVE"] = 0
    statistics["TRUE_NEGATIVE"] = 0

    manual_data_dicts = {}
    testing_data_dicts = {}
    with filepath.open(mode="r") as filehandle:
        csv_reader = csv.reader(filehandle, delimiter=",", quotechar='"')

        for row in csv_reader:
            # for manual_data
            if row[1] == "manual":
                data_dict = manual_data_dicts.get(row[0], dict())
                data_dict[row[2]] = row[3]
                manual_data_dicts[row[0]] = data_dict
            elif row[1] == testing_data_title:
                data_dict = testing_data_dicts.get(row[0], dict())
                data_dict[row[2]] = row[3]
                testing_data_dicts[row[0]] = data_dict

    for strain in GENOME_LENGTHS:
        manual_data_dict = manual_data_dicts.get(strain, dict())
        testing_data_dict = testing_data_dicts.get(strain, dict())

        for prophage_id, ends_tuple in manual_data_dict.items():
            ends_tuple = eval(ends_tuple)
            # check forward or reverse
            if ends_tuple[1] - ends_tuple[0] < 0:
                length_manual += ends_tuple[0] - ends_tuple[1]
            else:
                length_manual += ends_tuple[1] - ends_tuple[0]

        for prophage_id, ends_tuple in testing_data_dict.items():
            test_ends = eval(ends_tuple)
            if test_ends[1] - test_ends[0] < 0:
                length_test += test_ends[0] - test_ends[1]
            else:
                length_test += test_ends[1] - test_ends[0]

        per_phage = stats(manual_data_dict, testing_data_dict)
        true_positive += per_phage.get("TRUE_POSITIVE")
        false_positive += per_phage.get("FALSE_POSITIVE")
        false_negative += per_phage.get("FALSE_NEGATIVE")

        per_phage["TRUE_NEGATIVE"] = (GENOME_LENGTHS[strain] -
                                      length_manual - false_positive)
        true_negative += per_phage.get("TRUE_NEGATIVE")

        if per_strain:
            per_strain_metrics = metrics(per_phage)

            strain_title = ": ".join([testing_data_title, strain])
            print_data(per_phage, per_strain_metrics, strain_title)

    statistics["TRUE_POSITIVE"] += true_positive
    statistics["FALSE_POSITIVE"] += false_positive
    statistics["FALSE_NEGATIVE"] += false_negative
    statistics["TRUE_NEGATIVE"] += true_negative
    return statistics


def metrics(stats):
    """Calculate the metric data.

    :param stats: statistics data with tp, tn, fn, fp values
    :type stats: dict
    :return: Dictionary containing sensitivty, positive predictive value,
             accuracy and Matthews Correlation Coefficient (MCC)
    :rtype: dict
    """
    # collect data
    tp = stats["TRUE_POSITIVE"]
    tn = stats["TRUE_NEGATIVE"]
    fp = stats["FALSE_POSITIVE"]
    fn = stats["FALSE_NEGATIVE"]

    # calculate metric values
    if (tp + fn) == 0:
        sn = 1
    else:
        sn = (tp/(tp+fn))*100

    if tp + fp == 0:
        ppv = 1
    else:
        ppv = (tp/(tp+fp))*100

    if (tp+tn+fp+fn) == 0:
        acc = 1
    else:
        acc = ((tp+tn)/(tp+tn+fp+fn))*100

    confusion_matrix = (tp+fp)*(tp+fn)*(tn+fp)*(tn+fn)
    if confusion_matrix == 0:
        confusion_matrix = 1

    mcc = ((tn*tp) - (fn*fp))/math.sqrt(confusion_matrix)

    # return metrics as dictionary
    return {"sensitivity": round(sn, 3),
            "ppv": round(ppv, 3),
            "accuracy": round(acc, 3),
            "mcc": round(mcc, 3)}


def get_child_ends(prophage_filepath, parent_contigs):
    """Return the ends of the prophage region as a tuple.

    :param prophage_filepath: Filepath to the hypothetical prophage FASTA
    :type prophage_filepath: Path
    :param parent_seq: Sequence of the parent genome
    :return: Tuple of prophage ends
    :rtype: tuple
    """
    # read all the prophages in that folder

    child_contigs = open_file(prophage_filepath)
    child_tuples = []

    for child_seq in child_contigs:
        for parent_seq in parent_contigs:
            # find the start of the prophage regions - contigs
            start = find_sub(parent_seq, child_contigs)

            # not found
            if start is None:
                continue

            # tuple with ends (stop = start+length)
            child_tuples.append((start, start+len(child_seq)))

    return child_tuples


def comparison(dir, manual_name, software_name, per_strain=False):
    """Compare the data.

    :param dir: Path to working directory
    :type dir: Path
    :param manual_name: Name of directory with manually annotated genomes
    :type manual_name: str
    :param software_name: Name of software being tested
    :type software_name: str
    """
    # user enters the name of the folder with manually annotated
    # and the software name
    path_to_manual = dir/"prophages"/manual_name
    path_to_software = dir/"prophages"/software_name

    genome_path = dir/"genomes"

    csv_filename = software_name + "_testing_results.csv"

    csv_path = dir/csv_filename

    with open(csv_path, mode="w") as csv_file:

        writer = csv.DictWriter(csv_file, fieldnames=CSV_HEADERS)
        writer.writeheader()

        for genome in sorted(genome_path.iterdir()):
            strain_name = genome.stem
            ref_seq = open_file(genome, True)     # reference genome sequence
            manual_strain_path = path_to_manual/strain_name
            software_strain_path = path_to_software/strain_name

            # data to csv is written as dictionaries with headers as keys
            if manual_strain_path.is_dir():
                for prophage in sorted(manual_strain_path.iterdir()):
                    manual_ends = get_child_ends(prophage, ref_seq)
                    if manual_ends is None:
                        # this should not happen
                        print("ERROR: Prophage not found: ", prophage)
                        dict = {"strain": strain_name,
                                "software": "manual",
                                "prophage": prophage.stem,
                                "ends": "NONE"}
                        writer.writerow(dict)
                    else:
                        # data written for each manually annotated genome
                        for ends_tuple in manual_ends:
                            dict = {"strain": strain_name,
                                    "software": "manual",
                                    "prophage": prophage.stem,
                                    "ends": ends_tuple}
                            writer.writerow(dict)

            if software_strain_path.is_dir():
                for prophage in sorted(software_strain_path.iterdir()):
                    # check that no hidden files are being read in
                    # only read in fasta files
                    if prophage.suffix == ".fasta":
                        software_ends = get_child_ends(prophage, ref_seq)
                    else:
                        continue

                    # same logic as above for software data
                    if software_ends is None:
                        # if the prophage found by the software not found in
                        # host this should not happen
                        dict = {"strain": strain_name,
                                "software": software_name,
                                "prophage": prophage.stem,
                                "ends": "NONE"}
                        writer.writerow(dict)
                    else:
                        # write software data to the csv
                        for ends_tuple in software_ends:
                            dict = {"strain": strain_name,
                                    "software": software_name,
                                    "prophage": prophage.stem,
                                    "ends": ends_tuple}
                            writer.writerow(dict)
        csv_file.close()    # close csv file

        # collect the statistics
        stats = collect_stats(csv_path, per_strain=per_strain)
        metric_data = metrics(stats)        # metrics

        # print all the relevant data
        print_data(stats, metric_data, software_name)


def print_data(stats, metrics, title):
    """Print the stats and metric data.

    :param stats: Statistics for the software
    :type stats: dict
    :param metrics: Metrics for the software
    :type metrics: dict
    """
    print("\n\n")
    print("==============================================================")
    print(f"{title}:")
    print("--------------------------------------------------------------")
    print("TRUE POSITIVE\tFALSE POSITIVE\tTRUE NEGATIVE\tFALSE NEGATIVE")
    print("--------------------------------------------------------------")
    print(f"{stats['TRUE_POSITIVE']}\t\t{stats['FALSE_POSITIVE']}\t",
          f"\t{stats['TRUE_NEGATIVE']}\t{stats['FALSE_NEGATIVE']}")

    print("\n\n-----------------------------------------------------------")
    print("Sensitivity\tPPV\t\tAccuracy\tMCC")
    print("-----------------------------------------------------------")
    print(f"{metrics['sensitivity']}\t",
          f"\t{metrics['ppv']}\t\t{metrics['accuracy']}\t\t{metrics['mcc']}")
    print("==============================================================")


def parse_args(unparsed_args):
    """Parse commandline arguments.

    :return: Working directory where the sequences are located
    :rtype: Path
    """
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("dir", type=pathlib.Path,
                        help="path to working directory")
    parser.add_argument("manual", type=str,
                        help="name of directory with manually annotated "
                             "prophages")
    parser.add_argument("software", type=str,
                        help="name of the software being tested")

    parser.add_argument("--per_strain", action="store_true",
                        help="print metrics per genome")

    args = parser.parse_args(unparsed_args)

    return args


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    # run comparison
    comparison(args.dir, args.manual, args.software,
               per_strain=args.per_strain)


if __name__ == "__main__":
    main()



import pandas as pd
import plotly.express as px


blue, red = "#0008ff", "#ff0800"
colormap = {"prophage": blue, "bacteria": red}


class ProphageClassifier:
    """
    Classifier that uses a Naive Bayes-type strategy to construct
    probability distributions for an n-feature space
    """
    def __init__(self):
        """
        Constructor for instances of ProphageClassifier.
        """
        self.distributions_ = dict()

    @property
    def n_features(self):
        return len(self.distributions_)

    def fit(self, x, y, plot=False):
        """
        Fits a model that attempts to differentiate between

        :param x: training data to fit a model against
        :type x: pandas.DataFrame
        :param y: training data class labels
        :type y: list or numpy.ndarray or pd.DataFrame
        :param plot: show distribution of feature space?
        :type plot: bool
        """
        # Build a probability distribution for each feature
        prophage_df = x.iloc[[i for i, value in enumerate(y) if value == 1], :]
        bacteria_df = x.iloc[[i for i, value in enumerate(y) if value == 0], :]

        ratio = float(len(bacteria_df)) / len(prophage_df)

        figs = list()
        for i, feature in enumerate(x.columns):
            prophage_series = prophage_df.loc[:, feature]
            bacteria_series = bacteria_df.loc[:, feature]

            prophage_hist = Histogram(prophage_series)
            bacteria_hist = Histogram(bacteria_series, prophage_hist.bin_width)

            if plot:
                pdf = prophage_hist.as_dataframe()
                pdf["class"] = ["prophage"] * len(pdf)
                pdf["values"] = [float(x)/prophage_hist.n_samples * 100 for x
                                 in pdf["values"]]
                pdf["color"] = [blue] * len(pdf)

                bdf = bacteria_hist.as_dataframe()
                bdf["class"] = ["bacteria"] * len(bdf)
                bdf["values"] = [float(x)/bacteria_hist.n_samples * 100 for x
                                 in bdf["values"]]
                bdf["color"] = [red] * len(bdf)

                df = pd.concat([pdf, bdf])
                fig = px.bar(data_frame=df, x="bins", y="values",
                             color="class", barmode="overlay",
                             color_discrete_map=colormap,
                             template="simple_white")
                figs.append((feature, fig))

            dist = ProbabilityDistribution(prophage_hist, bacteria_hist,
                                           weights=[1, ratio])

            self.distributions_[feature] = dist

        return figs

    def predict_proba(self, x, feature_weights=None):
        """
        Uses the known probability distributions for input features
        that each input row belongs to a prophage.

        NOTE: sum(feature_weights) must equal 1.

        :param x: the data rows to make predictions for
        :type x: pandas.DataFrame
        :param feature_weights: how much weigh to place on each feature
        :type feature_weights: list of float
        :return: temp_proba
        """
        if len(x.columns) != self.n_features:
            raise ValueError(f"input dataframe must have {self.n_features} "
                             f"columns")

        if feature_weights and len(feature_weights) != self.n_features:
            raise ValueError(f"feature_weights should have {self.n_features} "
                             f"elements")

        if feature_weights and sum(feature_weights) != 1:
            raise ValueError("feature_weights must sum to 1")

        if not feature_weights:
            feature_weights = [1.0 / self.n_features] * self.n_features

        temp_proba = list()
        for i, row in x.iterrows():
            probabilities = list()
            for feature in row.keys().values:
                value = row[feature]
                proba = self.distributions_[feature].get_probability(value)
                probabilities.append(proba)
            row_proba = 0
            for weight, proba in zip(feature_weights, probabilities):
                row_proba += weight * proba
            temp_proba.append(row_proba)

        return temp_proba

    def predict(self, x, feature_weights=None, alpha=0.5):
        """
        Get binary class predictions by calling predict_proba and
        setting values to 0 if < alpha, else 1.

        :param x: the data rows to make predictions for
        :type x: pandas.DataFrame
        :param feature_weights: how much weigh to place on each feature
        :type feature_weights: list of float
        :param alpha: threshold for class 1
        :type alpha: float
        :return: temp_predict
        """
        proba = self.predict_proba(x, feature_weights)
        temp_predict = list()
        for p in proba:
            if p >= alpha:
                temp_predict.append(1)
            else:
                temp_predict.append(0)

        return temp_predict

    def __len__(self):
        return len(self.distributions_)


class Histogram:
    def __init__(self, data, bin_width=None):
        """
        Constructor for instances of the Histogram class.

        :param data: data to build the histogram from
        :type data: pandas.Series
        :param bin_width: the desired histogram bin width
        :type bin_width: int or float
        """
        self.hist = dict()
        self.n_samples = len(data)

        if bin_width:
            self.bin_width = bin_width
        else:
            data_range = data.max() - data.min()

            if 0 < data_range <= 1:         # For very small numbers
                self.bin_width = 0.01
            elif 0 < data_range <= 100:     # For smallish numbers
                self.bin_width = 1
            else:                           # Everything else
                self.bin_width = 10

        for value in data:
            index = round(value // self.bin_width * self.bin_width, 2)
            if index in self.hist:
                self.hist[index] += 1
            else:
                self.hist[index] = 1

    def as_dataframe(self):
        return pd.DataFrame({"bins": self.hist.keys(),
                             "values": self.hist.values()})


class ProbabilityDistribution:
    def __init__(self, prophage_hist, bacteria_hist, weights=None):
        """

        :param prophage_hist: prophage histogram
        :type prophage_hist: Histogram
        :param bacteria_hist: bacteria histogram
        :type bacteria_hist: Histogram
        :param weights: adjust weights to account for uneven sampling
        :type weights: list of int or list of float or None
        """
        if not prophage_hist.bin_width == bacteria_hist.bin_width:
            raise ValueError("input histograms must have the same bin_width")

        self.distribution = dict()

        self.minimum = None
        self.maximum = None

        self.bin_width = prophage_hist.bin_width

        if weights:
            prophage_weight, bacteria_weight = weights
        else:
            prophage_weight, bacteria_weight = 1, 1

        prophage_count = prophage_hist.n_samples * prophage_weight
        bacteria_count = bacteria_hist.n_samples * bacteria_weight

        temp_distribution = dict()

        for prophage_key, prophage_value in prophage_hist.hist.items():
            numer = float(prophage_value) / prophage_count * prophage_weight
            if prophage_key in bacteria_hist.hist:
                bacteria_value = bacteria_hist.hist[prophage_key]
                bacteria_proportion = float(bacteria_value) / bacteria_count
                bacteria_proportion *= bacteria_weight
            else:
                bacteria_proportion = 0.0
            denom = numer + bacteria_proportion
            prophage_probability = round(float(numer) / denom, 2)
            temp_distribution[prophage_key] = prophage_probability

        for bacteria_key, bacteria_value in bacteria_hist.hist.items():
            if bacteria_key not in temp_distribution:
                temp_distribution[bacteria_key] = 0.0

        sorted_keys = sorted(temp_distribution.keys())
        self.minimum, self.maximum = sorted_keys[0], sorted_keys[-1]
        for key in sorted_keys:
            self.distribution[key] = temp_distribution[key]

        self._smooth()

    def _smooth(self):
        expect_steps = (self.maximum - self.minimum + self.bin_width) // \
                       self.bin_width

        # Fill missing values with their left neighbor
        for i in range(round(expect_steps)):
            key = round(self.minimum + (i * self.bin_width), 2)
            if key not in self.distribution:
                prior_key = round(key - self.bin_width, 2)
                self.distribution[key] = self.distribution[prior_key]

    def get_probability(self, value):
        if value < self.minimum:
            probability = 1.0
        elif value > self.maximum:
            probability = 0.0
        else:
            key = round(value // self.bin_width * self.bin_width, 2)
            probability = self.distribution[key]

        return probability



"""Utility script to index protein-coding gene sequences.
"""
import argparse
import pathlib
import sys

from Bio import SeqIO

from depht.data import GLOBAL_VARIABLES
from depht_train.functions.fileio import (
    read_cluster_table_file, write_cluster_file)

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
TABLE = 11
NAME = GLOBAL_VARIABLES["sequences"]["name"]
DEFAULT_PRODUCT = GLOBAL_VARIABLES["sequences"]["default_product"]


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def parse_args(unparsed_args):
    """Parse commandline arguments

    :param unparsed_args: command line args
    :type unparsed_args: list[str]
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("input_dir", type=pathlib.Path)
    parser.add_argument("output_dir", type=pathlib.Path)

    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-n", "--name", type=str, default=NAME)

    parser.add_argument("-tt", "--translation_table", type=int,
                        default=TABLE)
    parser.add_argument("-cf", "--cluster_table", type=pathlib.Path,
                        default=None)

    args = parser.parse_args(unparsed_args)
    return args


def index_sequences(input_dir, output_dir,
                    name=NAME, table=TABLE,
                    default_product=DEFAULT_PRODUCT,
                    cluster_table=None):
    """Function to index protein coding features and related metadata.

    :param input_dir: Path to annotated GenBank-formatted genome sequences.
    :type input_dir: pathlib.Path
    :param output_dir: Path to write index output.
    :type output_dir: pathlib.Path
    :param name: Stem name of the outputted index file.
    :type name: str
    :param table: Translation table key to translate sequences, if necessary
    :type table: int
    :param default_product: Product to annotated unannotated sequences with.
    :type default_product: str
    :param cluster_table: Path to a csv table mapping genomes to clusters.
    :type cluster_table: pathlib.Path
    """
    cds_features = []
    for input_file in input_dir.iterdir():
        records = [record for record in SeqIO.parse(input_file, "gb")]

        for record in records:
            for index, feature in enumerate(record.features):
                if feature.type != "CDS":
                    continue

                # Checks to see if the CDS feature has an annotated translation
                translation = feature.qualifiers.get("translation")
                # If the feature lacks a translation qualifier, create one
                if not translation:
                    sequence = feature.extract(record.seq)
                    translation = sequence.translate(table=table, to_stop=True)

                    feature.qualifiers["translation"] = [translation]

                # Checks to see if the CDS feature has an annotated function
                product = feature.qualifiers.get("product")
                # If the feature lacks a function qualifier, create one
                if not product:
                    feature.qualifiers["product"] = [default_product]
                else:
                    # If the feature has a blank function qualifier, create one
                    if not product[0]:
                        feature.qualifiers["product"] = [default_product]

                # Checks to see if the CDS feature has an annotated locus tag
                locus_tag = feature.qualifiers.get("locus_tag")
                if not locus_tag:
                    feature.qualifiers["locus_tag"] = ["_".join(
                                                    [record.id, str(index+1)])]

                feature.qualifiers["note"] = [input_file.stem]

                cds_features.append(feature)

    output_dir.mkdir(exist_ok=True, parents=True)

    fasta_file = output_dir.joinpath(".".join([name, "fasta"]))
    index_file = output_dir.joinpath(".".join([name, "pgi"]))
    write_index_files(cds_features, index_file, fasta_file)

    cluster_file = None
    if cluster_table is not None:
        clustered_ids = get_clustered_records(cluster_table)

        cluster_file = output_dir.joinpath(".".join([name, "ci"]))
        write_cluster_file(clustered_ids, cluster_file)

    return fasta_file, index_file, cluster_file


def get_clustered_records(cluster_table_path):
    record_cluster_map = read_cluster_table_file(cluster_table_path)

    cluster_records_map = dict()
    for name, cluster in record_cluster_map.items():
        if cluster == "":
            cluster = None

        names = cluster_records_map.get(cluster, list())
        names.append(name)
        cluster_records_map[cluster] = names

    clustered_records = [names for c, names in cluster_records_map.items()
                         if c is not None]

    return clustered_records


def write_index_files(cds_features, index_file, fasta_file):
    """Function to write indexed sequences to file.

    :param cds_features: SeqFeature objects to be indexed.
    :type cds_features: list
    :param fasta_file: Path to write indexed sequences in fasta-format.
    :type fasta_file: pathlib.Path
    :param index_file: Path to write index and relevant metadata.
    :type index_file: pathlib.Path
    """
    with fasta_file.open(mode="w") as fasta_filehandle:
        with index_file.open(mode="w") as index_filehandle:
            for index, feature in enumerate(cds_features):
                locus_tag = feature.qualifiers["locus_tag"][0]
                translation = feature.qualifiers["translation"][0]
                product = feature.qualifiers["product"][0]
                parent = feature.qualifiers["note"][0]

                fasta_filehandle.write("".join([">", str(index), "\n"]))
                fasta_filehandle.write("".join([translation, "\n"]))

                index_filehandle.write("\t".join(
                                       [str(index), locus_tag, product,
                                        parent, "\n"]))


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    index_sequences(args.input_dir, args.output_dir, name=args.name,
                    table=args.translation_table,
                    cluster_table=args.cluster_table)


if __name__ == "__main__":
    main()



from depht.classes.hhresult import HHResult
from depht.data import PARAMETERS
from depht.functions.fasta import write_fasta
from depht.functions.multiprocess import parallelize
from depht.functions.subprocess import run_command

HHSEARCH_EVALUE = PARAMETERS["phage_homology_search"]["evalue"]
HHSEARCH_PROB = PARAMETERS["phage_homology_search"]["probability"]
HHSEARCH_COV = PARAMETERS["phage_homology_search"]["coverage"]


def hhsearch(query, outfile, db, evalue=HHSEARCH_EVALUE, prob=HHSEARCH_PROB,
             cov=HHSEARCH_COV):
    """
    Runs a single instance of hhsearch using the prescribed paths
    and e-value.

    :param query: the filepath to query against an HHSuite3 database
    :type query: pathlib.Path
    :param outfile: the desired output filepath
    :type outfile: pathlib.Path
    :param db: the HHSuite3 database to query
    :type db: pathlib.Path
    :param evalue: the e-value cutoff to use
    :type evalue: float
    :param prob: the probability cutoff to use
    :type prob: float
    :param cov: the coverage cutoff to use
    :type cov: float
    """
    command = f"hhsearch -i {query} -d {db} -o {outfile} -e {evalue} -E " \
              f"{evalue} -p {prob} -cov {cov} -cpu 1"
    run_command(command)


def find_single_homologs(header, sequence, db, tmp_dir, prob=HHSEARCH_PROB,
                         cov=HHSEARCH_COV):
    """
    Runs hhsearch to find functionally annotated homolog(s) for a
    single protein sequence in the indicated database.

    Returns a tuple containing the header and the label of the
    best-scoring match with probability greater than `prob`, if such a
    match exists.

    :param cov:
    :param header: a label for the sequence to search
    :type header: str
    :param sequence: the sequence to find homologs for
    :type sequence: str
    :param db: the database to find homologs in
    :type db: pathlib.Path
    :param tmp_dir: the directory where temporary files can go
    :type tmp_dir: pathlib.Path
    :param prob: probability threshold to keep a match
    :type prob: int or float
    :return: hhresult
    """
    # Set up file paths
    query_file = tmp_dir.joinpath(f"{header}.fasta")
    output_file = query_file.with_suffix(".hhr")

    # Write the fasta file, and run hhsearch
    write_fasta([header], [sequence], query_file)
    hhsearch(query_file, output_file, db)

    # Parse the hhsearch result
    hhresult, best_match, best_probability = HHResult(output_file), None, 0.0
    hhresult.parse_result()

    # Cull low-probability and low-coverage matches
    keep = list()
    for match in hhresult.matches:
        probability = float(match.probability) >= prob
        qry_cov = 100 * float(match.match_cols) / len(sequence) >= cov
        sbj_cov = 100 * float(match.match_cols) / float(match.hit_length) >= cov

        if probability and qry_cov and sbj_cov:
            keep.append(match)
    hhresult.matches = keep
                             
    if hhresult.matches:
        # Best match is the one with the highest bit-score
        hhresult.matches.sort(key=lambda x: float(x.score), reverse=True)
        best_match = hhresult.matches[0].target_id
        best_probability = hhresult.matches[0].probability
    else:
        if output_file.is_file():
            output_file.unlink()

    query_file.unlink()

    return header, best_match, best_probability


def find_batch_homologs(headers, sequences, db, tmp_dir, cpus):
    """
    Runs hhsearch to find functionally annotated homologs for each
    of the given translations in the indicated database.

    :param headers: labels for the sequence to search
    :type headers: list of str
    :param sequences: the sequences to find homologs for
    :type sequences: list of str
    :param db: the database to find homologs in
    :type db: pathlib.Path
    :param tmp_dir: the directory where temporary files can go
    :type tmp_dir: pathlib.Path
    :param cpus: how many CPU cores to use?
    :type cpus: int
    :return: homologs
    """
    jobs = list()
    for header, sequence in zip(headers, sequences):
        jobs.append((header, sequence, db, tmp_dir))
    homologs = parallelize(jobs, cpus, find_single_homologs)

    return homologs


def find_homologs(contigs, prophage_coords, db, tmp_dir, cpus, min_length=150,
                  cache_scores=True):
    """
    Convenience function for finding all prophage-predicted gene homologs
    across all contigs of a genome.

    Updates contig features in-place.

    :param contigs: the contigs from an input file
    :type contigs: list of Bio.SeqRecord.SeqRecord
    :param prophage_coords: predicted prophage coords on each contig
    :type prophage_coords: list of list of tuple(int, int)
    :param db: the database to find homologs in
    :type db: pathlib.Path
    :param tmp_dir: the directory where temporary files can go
    :type tmp_dir: pathlib.Path
    :param cpus: how many CPU cores to use?
    :type cpus: int
    :param min_length: don't find homologs for genes shorter than this
    :type min_length: int
    :param cache_scores: Toggles whether to store probabilities in the contig
    :type cache_scores: bool
    """
    # Iterate over contigs and prophage coordinate predictions together
    for contig, contig_prophage_coords in zip(contigs, prophage_coords):
        map_geneid_to_feature = dict()
        batch_geneids, batch_sequences = list(), list()
        for i, feature in enumerate(contig.genes):
            geneid = contig.gene_ids[i]

            translation = feature.qualifiers["translation"][0]

            # Only hhsearch if length > min_length
            if len(translation) < min_length:
                continue

            # Only check features that overlap or are in prophages or haven't been
            # searched before
            search = False
            if __feature_in_prophage(feature, contig_prophage_coords):
                search = True
                if contig.hhsearch_scores:
                    if contig.hhsearch_scores[i] > 0:
                        search = False

            if search:
                map_geneid_to_feature[geneid] = feature
                batch_geneids.append(geneid)
                batch_sequences.append(translation)

        homologs = find_batch_homologs(
            batch_geneids, batch_sequences, db, tmp_dir, cpus)

        hhsearch_scores = [0.0] * len(contig.gene_ids)
        for geneid, product, prob in homologs:
            if product:
                feature = map_geneid_to_feature[geneid]
                feature.qualifiers["product"] = [product]

                hhsearch_scores[contig.gene_ids.index(geneid)] = float(prob)

        if cache_scores:
            contig.update_hhsearch_scores(hhsearch_scores)


def __feature_in_prophage(feature, contig_prophage_coords):
    """
    Checks whether the indicated feature overlaps the coordinates of
    a predicted prophage.

    :param feature: the feature to check
    :type feature: Bio.SeqFeature.SeqFeature
    :param contig_prophage_coords: predicted prophage coordinates
    :type contig_prophage_coords: list of tuple(int, int)
    """
    for start, end in contig_prophage_coords:
        if (start < feature.location.start < end) or \
                (start < feature.location.end < end):
            return True
    return False



import pathlib


def parse_fasta(filepath):
    """
    Parses a fasta file into a list of headers and a corresponding list
    of sequences.

    :param filepath: fasta file to parse
    :type filepath: pathlib.Path or str
    :return: headers, sequences
    """
    headers, sequences = list(), list()

    fasta_reader = open(filepath, "r")

    cache = list()
    for line in fasta_reader:
        # If header line, flush the cache and store header
        if line.startswith(">"):
            sequences.append("".join(cache))
            cache = list()
            headers.append(line.lstrip(">").rstrip())
        # Otherwise, append to the cache
        else:
            cache.append(line.rstrip())

    # Flush the last sequence out of the cache, and pop empty sequence
    sequences.append("".join(cache))
    sequences.pop(0)

    # Close the file handle
    fasta_reader.close()

    return headers, sequences


def write_fasta(headers, sequences, filepath, width=80):
    """
    Writes the given headers and sequences to filepath.

    :param headers: the sequence labels
    :type headers: list of str
    :param sequences: the sequences to write
    :type sequences: list of str
    :param filepath: the fasta file to write
    :type filepath: pathlib.Path or str
    :param width: maximum number of characters per sequence line
    :type width: int
    """
    if not isinstance(headers, list) or not isinstance(sequences, list):
        raise TypeError(f"headers and sequences should be lists of strings")

    fasta_writer = open(filepath, "w")

    for header, sequence in zip(headers, sequences):
        fasta_writer.write(f">{header}\n")
        # Use string slicing to split the sequence to satisfy width param
        for i in range(0, len(sequence), width):
            fasta_writer.write(f"{sequence[i:i + width]}\n")

    # Close the file handle
    fasta_writer.close()



"""Wrapper around a subset of the subprocess and shlex built-in libraries
to make running subprocesses easy."""

import shlex
from subprocess import Popen, PIPE


def run_command(command):
    """Run the indicated command as a subprocess.

    If `verbose` is True, stdout and stderr will be returned and the
    code block that called this one can decide how to handle it.

    If the command invokes a program that is not globally executable,
    it will raise a FileNotFoundError with a message like "No such
    file or directory: 'xxx'".

    :param command: the command to run
    :type command: str
    :raises: FileNotFoundError
    :return: out, err
    """
    command = shlex.split(command)

    # Reading from stdout/stderr is blocking, no need to call sp.wait()
    with Popen(command, stdout=PIPE, stderr=PIPE, close_fds=True) as process:
        out = process.stdout.read().decode("utf-8")
        err = process.stderr.read().decode("utf-8")

    return out, err



import numpy as np


class Database:
    """
    A Database is a collection of geneids and translations, with methods
    for conveniently accessing data keyed on either translations or geneids,
    and for removing redundancy for faster processing.
    """
    def __init__(self, geneids=tuple(), translations=tuple()):
        """
        Database constructor - assumes that geneids and translations are
        index-paired: e.g. geneids[i] is the label for the sequence at
        translations[i].

        :param geneids: gene labels
        :type geneids: list of str or tuple of str
        :param translations: gene sequences
        :type translations: list of str or tuple of str
        :return: database
        :rtype: Database
        """
        self.all_genes = dict()
        self.non_redundant_genes = dict()
        self.add_genes(geneids, translations)

    def add_genes(self, geneids, translations):
        """
        Database batch submission tool - assumes that geneids and
        translations are index-paired: e.g. geneids[i] is the label for
        the sequence at translations[i].

        Iterates over paired elements and calls `add_gene()` on them.

        :param geneids: gene labels to add
        :type geneids: list or tuple
        :param translations: gene sequences to add
        :type translations: list or tuple
        :return:
        """
        if len(geneids) != len(translations):
            raise ValueError(f"number of geneids ({len(geneids)}) != number of translations ({len(translations)}")

        for geneid, translation in zip(geneids, translations):
            self.add_gene(geneid, translation)

    def add_gene(self, geneid, translation):
        """
        Database single submission tool - assumes that this pair of
        geneid and translation are not already present in the database.

        :param geneid: gene label to add
        :type geneid: str
        :param translation: gene sequence to add
        :type translation: str
        :return:
        """
        if geneid in self.all_genes.keys():
            raise ValueError(f"{geneid} already in database")

        self.all_genes[geneid] = translation
        non_redundant_group = self.non_redundant_genes.get(translation, list())
        non_redundant_group.append(geneid)
        self.non_redundant_genes[translation] = non_redundant_group

    def get_geneids_from_translation(self, translation):
        """
        Returns the list of geneids associated with a translation.

        :param translation: the gene sequence to get labels for
        :return: geneids
        """
        return self.non_redundant_genes[translation]

    def get_translation_from_geneid(self, geneid):
        """
        Returns the translation associated with a geneid.

        :param geneid: the gene label to get a translation for
        :return: translation
        """
        return self.all_genes[geneid]

    def get_non_redundant_array(self):
        temp_list = list()
        max_len = 0
        for translation in self.non_redundant_genes.keys():
            length = len(translation)
            if length > max_len:
                max_len = length
            temp_list.append((translation, length))

        datatype = np.dtype([("Translation", np.unicode, max_len), ("Length", "u4")])
        return np.sort(np.array(temp_list, dtype=datatype), order=["Length", "Translation"])

    def __repr__(self):
        s = [f">{geneids[0]}\n{translation}\n" for translation, geneids in self.non_redundant_genes.items()]
        return "".join(s)

    def __str__(self):
        s = [f">{geneid}\n{translation}\n" for geneid, translation in self.all_genes.items()]
        return "".join(s)

    def __len__(self):
        return len(self.all_genes)


class Pham(Database):
    """
    A Pham is a slightly extended Database implementation, whose main
    difference is that instances of pham are sortable (on length).
    """
    def __init__(self, geneids, translations):
        """
        Constructor for Pham

        :param geneids: the geneids of the sequences assorted into this Pham
        :type geneids: list of str
        :param translations: the sequences assorted into this Pham
        :type translations: list of str
        """
        super().__init__(geneids, translations)

    def get_translations(self):
        return self.non_redundant_genes.keys()

    def __lt__(self, other):
        return len(self) < len(other)



from dna_features_viewer import BiopythonTranslator

from depht.data import GLOBAL_VARIABLES

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_PRODUCT = GLOBAL_VARIABLES["sequences"]["default_product"]
DEFAULT_FONT_FAMILY = "monospace"


class CircularSourceFeatureTranslator(BiopythonTranslator):
    """Translator for Circular Graphic."""

    def compute_feature_color(self, feature):
        """Compute the color of the feature."""
        if feature.location.strand == -1:   # reverse
            source_color = "red"
        else:   # forward
            source_color = "#449f66"

        color_map = {
            "source": source_color,  # light orange
            "gene": "red",
            "tRNA": "yellow",
            "CDS": "red",  # pink
            "misc_recomb": "darkblue",
            "misc_feature": "#d1e9f1",  # light blue
            "backbone": "darkblue",
        }
        return color_map[feature.type]

    def compute_feature_label(self, feature):
        """Compute the label of the feature."""
        if feature.type not in ("source"):
            return None
        else:
            return feature.qualifiers.get("locus_tag")[0]

    def compute_feature_fontdict(self, feature):
        """Compute a  font dict for this feature."""
        return {"family": DEFAULT_FONT_FAMILY}

    def compute_feature_box_linewidth(self, feature):
        """Compute a box_linewidth for this feature."""
        return 1.0


class LinearFeatureTranslator(BiopythonTranslator):
    """Custom Translator adapted from BiopythonTranslator."""

    # Label fields indicates the order in which annotations fields are
    # considered to determine the feature's label

    # MODIFICATIONS:
    # label based on feature type!
    # display CDS unless HP, then --> gene
    # add display for tRNA, tmRNA - display note field
    # misc_recomb -- feature type - show sequence

    # bird's eye view of entire genome - circular
    # display - source?

    # feature.location - +/- for forward or reverse
    # feature.location.strand

    hyp = ["Hypothetical Protein", "hypothetical protein"]
    att_types = ["attL", "attR"]
    ignored_features_types = ["source", "gene"]
    label_fields = ["product", "name", "note", "gene"]

    def compute_feature_box_linewidth(self, feature):
        """Compute a box_linewidth for this feature."""
        return 1.0

    def compute_feature_linewidth(self, feature):
        """Compute the edge width of the feature's arrow/rectangle."""
        return 10

    def compute_feature_label(self, feature):
        """Compute the label of the feature."""
        label = feature.type
        for key in self.label_fields:
            if key in feature.qualifiers and len(feature.qualifiers[key]):
                label = feature.qualifiers[key]
                break

        if isinstance(label, list):
            label = "|".join([str(x) for x in label])

        if label in self.hyp:
            label = None
        elif feature.type == "misc_recomb":
            for att_type in self.att_types:
                if att_type in label:
                    label = att_type
                    break

        return label

    def compute_feature_html(self, feature):
        """Compute the tooltip display text of the feature"""
        label = self.compute_feature_label(feature)
        properties = {"gb_type": feature.type}

        if feature.type == "CDS":
            gene = str(feature.qualifiers["gene"][0])
            translation = feature.qualifiers["translation"][0]
            if label is None:
                label = DEFAULT_PRODUCT

            properties["gene"] = gene
            properties["product"] = label
            properties["translation"] = translation

        elif feature.type == "misc_recomb":
            sequence = feature.qualifiers["note"][0]

            properties["name"] = label
            properties["sequence"] = sequence

        elif feature.type in ("tRNA", "tmRNA"):
            if label is None:
                label = ""

            properties["product"] = label
            properties["codon"] = feature.qualifiers.get("note", [""])[0]

        return properties

    def compute_feature_fontdict(self, feature):
        """Compute a  font dict for this feature."""
        return {"family": DEFAULT_FONT_FAMILY}

    def compute_feature_label_link_color(self, feature):
        """Compute the color of the  line linking the label to its feature"""
        return None

    def compute_feature_legend_text(self, feature):
        """Compute the feature legend text."""
        return feature.type

    def compute_feature_box_color(self, feature):
        """Compute a box_color for this feature."""
        return "black"

    def compute_feature_color(self, feature):
        """Compute color for the given feature - differences based on type."""
        if feature.location.strand == 1:
            # green for forward
            CDS_color = "#449f66"
        else:
            # red for reverse
            CDS_color = "red"

        color_dict = {"source": "#34ff56",
                      "gene": "#f45562",
                      "CDS": CDS_color,
                      "misc_recomb": "#fbf3f6",
                      "tRNA": '#4554f4',
                      "tmRNA": '#56f443'}

        return color_dict[feature.type]



"""
DEPhT scans bacterial genomes looking for prophages. Regions
identified as prophage candidates are further scrutinized, and
attachment sites identified as accurately as possible before
prophage extraction and generating the final report.
"""
import argparse
import csv
import pathlib
import shutil
import sys
from datetime import datetime
from tempfile import mkdtemp

from Bio import SeqIO

from depht.classes.contig import Contig
from depht.classes.prophage import Prophage
from depht.data import GLOBAL_VARIABLES, PARAMETERS
from depht.functions.annotation import (annotate_record,
                                        cleanup_flatfile_records)
from depht.functions.att import find_attachment_site
from depht.functions.find_homologs import find_homologs
from depht.functions.mmseqs import assemble_bacterial_mask
from depht.functions.multiprocess import CPUS
from depht.functions.prophage_prediction import predict_prophage_coords
from depht.functions.sniff_format import sniff_format
from depht.functions.visualization import draw_complete_diagram

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEPHT_DIR = pathlib.Path().home().joinpath(
                                GLOBAL_VARIABLES["model_storage"]["home_dir"])
if not DEPHT_DIR.is_dir():
    DEPHT_DIR.mkdir()

MODEL_DIR = DEPHT_DIR.joinpath(GLOBAL_VARIABLES["model_storage"]["model_dir"])
if not MODEL_DIR.is_dir():
    MODEL_DIR.mkdir()

LOCAL_MODELS = [x.name for x in MODEL_DIR.iterdir() if x.name != ".DS_Store"]
if len(LOCAL_MODELS) == 0:
    # TODO: automatically retrieve models from OSF if the user says we can
    print("No DEPhT models found in ~/.depht/models. Please create one using "
          "'depht_train', or download models from https://osf.io/zt4n3/")
    sys.exit(0)

# Where temporary file I/O should happen
TMP_DIR = DEPHT_DIR.joinpath(GLOBAL_VARIABLES["model_storage"]["tmp_dir"])

# Model can't scan contigs with fewer CDS than window size
MIN_CDS_FEATURES = PARAMETERS["prophage_prediction"]["window"]

# For naming any identified prophages
PROPHAGE_PREFIX = GLOBAL_VARIABLES["phage_sequences"]["prophage_prefix"]
PROPHAGE_DELIMITER = GLOBAL_VARIABLES["phage_sequences"]["prophage_delimiter"]

# For attachment site detection
EXTEND_BY = PARAMETERS["att_detection"]["extention_length"]
ATT_SENSITIVITY = PARAMETERS["att_detection"]["att_sensitivity"]

# For deciding whether to cull predicted "prophages"
MIN_LENGTH = PARAMETERS["annotation"]["min_length"]
MIN_PRODUCTS_NORMAL = PARAMETERS["phage_homology_search"]["min_products"]
MIN_PRODUCTS_STRICT = PARAMETERS["phage_homology_search"][
                                 "min_products_strict"]

ANNOTATIONS = GLOBAL_VARIABLES["sequences"]["annotations"]
CONTIG_DATA_HEADER = ["Gene ID", "Start", "End", "Prediction",
                      "Bacterial Homology", "Phage Homology"]


def parse_args():
    """
    Parse command line arguments.

    :return: parsed_args
    """
    p = argparse.ArgumentParser(description=__doc__, prog="depht")
    # Positional arguments
    p.add_argument("infile", type=pathlib.Path, nargs="+",
                   help="path to genome file(s) to scan for prophages")
    p.add_argument("outdir",  type=pathlib.Path,
                   help="path where output files can be written")

    # Optional arguments
    p.add_argument("--model", type=str,
                   choices=set(LOCAL_MODELS), default=LOCAL_MODELS[0],
                   help=f"which local model should be used [default: "
                        f"{LOCAL_MODELS[0]}]")
    p.add_argument("-c", "--cpus",
                   metavar="", default=CPUS, type=int,
                   help=f"number of CPU cores to use [default: "
                        f"{CPUS}]")
    p.add_argument("-n", "--no-draw", action="store_false", dest="draw",
                   help="don't draw genome diagram for identified prophage(s)")
    p.add_argument("-m", "--mode",
                   choices=("fast", "normal", "sensitive"), default="normal",
                   help="select a runmode that favors speed or accuracy")
    p.add_argument("-s", "--att_sens", type=float,
                   default=ATT_SENSITIVITY, metavar="",
                   help="sensitivity parameter for att site detection.")
    p.add_argument("-d", "--dump-data", action="store_true",
                   help="dump all support data to outdir")
    p.add_argument("-v", "--verbose",
                   action="store_true",
                   help="print progress messages as the program runs")
    p.add_argument("-t", "--tmp-dir",
                   type=pathlib.Path, default=TMP_DIR, metavar="",
                   help=f"temporary directory to use for file I/O "
                        f"[default: {TMP_DIR}]")
    p.add_argument("-p", "--products",
                   type=int, default=None, metavar="",
                   help=("minimum number of phage homologs "
                         "to report a prophage"))
    p.add_argument("-l", "--length",
                   type=int, default=MIN_LENGTH, metavar="",
                   help=f"minimum length to report a prophage "
                        f"[default: {MIN_LENGTH}]")

    return p.parse_args()


def main():
    """
    Main function that interfaces with command line args and the
    program workflow.
    """
    # Invoke the help menu if no args given
    if len(sys.argv) == 1:
        sys.argv.append("-h")

    args = parse_args()

    infiles = args.infile       # What input file(s) did the user give?
    model = args.model          # Which model was selected?
    draw = args.draw            # Are we going to draw genome diagrams?
    dump = args.dump_data       # Are we dumping data into outdir when done?
    runmode = args.mode         # What runmode are we using?
    verbose = args.verbose      # Print verbose outputs?
    cpus = args.cpus            # How many physical cores can we use?
    att_sens = args.att_sens    # What's the att sensitivity modifier?
    min_length = args.length    # Minimum prophage length in bp

    # Get output dir and make sure it's a valid path
    outdir = pathlib.Path(args.outdir).resolve()
    if not outdir.is_dir():
        print(f"'{str(outdir)}' does not exist - creating it...")
        outdir.mkdir(parents=True)

    # Get the temporary directory, refresh it if it exists
    tmpdir = pathlib.Path(args.tmp_dir).resolve()
    if not tmpdir.is_dir():
        tmpdir.mkdir(parents=True)

    # Set up model directories
    model_dir = MODEL_DIR.joinpath(model)

    shell_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["shell_db"]["dir_name"])
    reference_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["reference_db"]["dir_name"])
    phage_homologs_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["phage_homologs"]["dir_name"])

    bact_ref_fasta = shell_db_dir.joinpath(
                        GLOBAL_VARIABLES["shell_db"]["fasta_name"])
    bact_ref_values = shell_db_dir.joinpath(
                        GLOBAL_VARIABLES["shell_db"]["hex_value_name"])
    essential_db = phage_homologs_dir.joinpath(
                        GLOBAL_VARIABLES["phage_homologs"]["essential_name"])
    extended_db = phage_homologs_dir.joinpath(
                        GLOBAL_VARIABLES["phage_homologs"]["extended_name"])
    blast_db = reference_db_dir.joinpath(
                        GLOBAL_VARIABLES["reference_db"]["name"])
    classifier = model_dir.joinpath(
                        GLOBAL_VARIABLES["classifier"]["name"])

    # Mark program start time
    mark = datetime.now()

    # OK, let's go!
    for infile in infiles:
        # Make sure filepath exists - skip if it doesn't
        if not infile.is_file():
            print(f"'{str(infile)}' does not exist - skipping it...")
            continue

        # Skip .DS_Store files
        if infile.name == ".DS_Store":
            print("skipping .DS_Store file...")
            continue

        # Set up a temporary directory for this genome
        genome_tmp_dir = pathlib.Path(mkdtemp(dir=tmpdir))
        if not genome_tmp_dir.is_dir():
            genome_tmp_dir.mkdir()

        # Automatically check the file format
        fmt = sniff_format(infile)

        # Parse all contigs of annotation-worthy length
        if verbose:
            print(f"\nparsing '{str(infile)}' as {fmt}...")

        # Parse the file and retain contigs above cutoff length
        records = [x for x in SeqIO.parse(infile, fmt) if len(x) >= MIN_LENGTH]

        if not records:
            if verbose:
                print(f"no {fmt}-formatted records of at least {MIN_LENGTH}bp "
                      f"found in '{str(infile)}' - skipping it...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        # Annotate contigs if format is "fasta"
        if fmt == "fasta":
            if verbose:
                print("annotating t(m)RNA and CDS genes de novo...")

            annotate_dir = genome_tmp_dir.joinpath("annotate")
            if not annotate_dir.is_dir():
                annotate_dir.mkdir()

            for record in records:
                annotate_record(record, annotate_dir)

        else:
            if verbose:
                print("using flat file annotation...")

            cleanup_flatfile_records(records)

        # Filter contigs that don't have enough CDS features
        records = [record for record in records if (len(
            [x for x in record.features if x.type == "CDS"]) >=
                                                    MIN_CDS_FEATURES)]

        if not records:
            print(f"no contigs with enough CDS features to analyze in "
                  f"'{str(infile)}' - skipping it...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        contigs = load_contigs(records)

        if verbose:
            print("masking conserved bacterial features...")

        mmseqs_dir = genome_tmp_dir.joinpath("mmseqs")
        if not mmseqs_dir.is_dir():
            mmseqs_dir.mkdir()

        # Detect conserved bacterial genes for each contig
        bacterial_masks = assemble_bacterial_mask(
            contigs, bact_ref_fasta, bact_ref_values, mmseqs_dir)

        if verbose:
            print("looking for high-probability prophage regions...")

        # Predict prophage coordinates for each contig
        prophage_predictions = list()
        for i, contig in enumerate(contigs):
            prediction = predict_prophage_coords(
                contig, classifier, EXTEND_BY, mask=bacterial_masks[i])

            filtered_prediction = []
            for pred in prediction:
                if len(range(*pred)) < min_length:
                    continue

                filtered_prediction.append(pred)

            prophage_predictions.append(filtered_prediction)

        if all([not any(x) for x in prophage_predictions]) and not dump:
            print(f"no complete prophages found in '{str(infile)}'...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        product_threshold = 0
        if runmode in ("normal", "sensitive"):
            hhsearch_dir = genome_tmp_dir.joinpath("hhsearch")
            if not hhsearch_dir.is_dir():
                hhsearch_dir.mkdir()

            # Search for phage gene remote homologs
            if verbose:
                print("searching for phage gene homologs...")

            find_homologs(contigs, prophage_predictions, essential_db,
                          hhsearch_dir, cpus)
            product_threshold = MIN_PRODUCTS_NORMAL

            if runmode == "sensitive":
                if verbose:
                    print("extending search for phage gene homologs...")
                find_homologs(contigs, prophage_predictions, extended_db,
                              hhsearch_dir, cpus, cache_scores=False)
                product_threshold = MIN_PRODUCTS_STRICT
        else:
            for contig in contigs:
                contig.fill_hhsearch_scores()

        if args.products is not None:
            product_threshold = args.products

        prophages = load_initial_prophages(contigs, prophage_predictions,
                                           product_threshold,
                                           prefix=PROPHAGE_PREFIX,
                                           delimiter=PROPHAGE_DELIMITER)

        if verbose and prophages:
            print("searching for attL/R...")

        # Set up directory where we can do attL/R detection
        att_dir = genome_tmp_dir.joinpath("att_core")
        if not att_dir.is_dir():
            att_dir.mkdir()

        # Detect attachment sites, where possible, for the predicted prophage
        search_space = att_sens * EXTEND_BY
        detect_att_sites(prophages, blast_db, search_space, att_dir)

        prophages = [prophage for prophage in prophages
                     if prophage.length >= min_length]

        if not prophages and not dump:
            print(f"no complete prophages found in '{str(infile)}'...")

            shutil.rmtree(genome_tmp_dir)  # clean up after ourselves
            continue

        if verbose:
            print("generating final reports...")

        genome_outdir = outdir.joinpath(f"{infile.stem}")
        if not genome_outdir.is_dir():
            genome_outdir.mkdir()

        draw_dir = genome_tmp_dir.joinpath("draw_diagram")
        if not draw_dir.is_dir():
            draw_dir.mkdir()
        write_prophage_output(genome_outdir, contigs, prophages, draw_dir,
                              draw)

        if dump:
            destination = genome_outdir.joinpath("tmp_data")

            if destination.exists():
                shutil.rmtree(destination)

            shutil.copytree(genome_tmp_dir, destination)
        shutil.rmtree(genome_tmp_dir)  # clean up after ourselves

    print(f"\nTotal runtime: {str(datetime.now() - mark)}")


# HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def load_contigs(contig_records):
    """Function to create Contig objects from bacterial sequence contig
    SeqRecords

    :param contig_records: Bacterial sequence contig SeqRecords
    :type contig_records: list
    :return: Contig options built from inputted SeqRecords
    :rtype: list
    """
    contigs = list()
    for record in contig_records:
        contig = Contig(record, record.name)
        contigs.append(contig)

    return contigs


def load_initial_prophages(contigs, prophage_predictions,
                           product_threshold=MIN_PRODUCTS_NORMAL,
                           prefix=PROPHAGE_PREFIX,
                           delimiter=PROPHAGE_DELIMITER):
    """Creates Prophage objects from initial prophage prediction coordinates
    and their respective parent SeqRecord objects.

    :param contigs: SeqRecord nucleotide sequence objects
    :type contigs: list[Contig]
    :param prophage_predictions: coordinates for predicted prophages
    :type prophage_predictions: list[list]
    :param product_threshold: number of products
    :type product_threshold:
    :param prefix: how should locus tags begin?
    :type prefix: str
    :param delimiter: how should locus tags be delimited
    :type delimiter: str
    :return: prophage objects that contain putative sequences and coordinates
    :rtype: list
    """
    prophages = []
    for contig_index, contig in enumerate(contigs):
        # Retrieve the contig SeqRecord associated with the coordinates
        contig_predictions = prophage_predictions[contig_index]

        prophage_index = 0
        for prophage_coordinates in contig_predictions:
            # Create a prophage ID from the SeqRecord ID
            prophage_id = "".join([prefix, contig.id,
                                   delimiter, str((prophage_index+1))])

            start = prophage_coordinates[0]
            if start < 0:
                start = 0

            end = prophage_coordinates[1]
            if end >= len(contig.seq):
                end = len(contig.seq) - 1

            prophage = Prophage(contig.record, prophage_id, start=start,
                                end=end)
            prophage.update()

            if len(prophage.products) < product_threshold:
                continue

            prophage_index += 1
            prophages.append(prophage)

    return prophages


def detect_att_sites(prophages, reference_db_path, search_space,
                     tmp_dir, min_kmer_score=5):
    """Detect attachment sites demarcating predicted prophage regions from
    the bacterial contig.

    :param prophages: Predicted prophages
    :type prophages: list
    :param reference_db_path: Path to the database with reference sequences
    :type reference_db_path: pathlib.Path
    :param search_space: Internal length of the prophage to check for att sites
    :type search_space: int
    :param tmp_dir: Path to place result files.
    :type tmp_dir: pathlib.Path
    :param min_kmer_score: Minimum length threshold of attachment sites.
    :type min_kmer_score: int
    """
    for prophage in prophages:
        # Create prophage working directory within temp directory
        working_dir = tmp_dir.joinpath(prophage.id)
        if not working_dir.is_dir():
            working_dir.mkdir()

        half_len = len(prophage.seq) // 2
        if search_space > half_len:
            search_space = half_len

        search_space = int(search_space)

        l_seq = str(prophage.seq[:search_space])
        r_seq = str(prophage.seq[-1*search_space:])

        l_name = f"{prophage.id}_left"
        r_name = f"{prophage.id}_right"

        att_data = find_attachment_site(
                                prophage, l_seq, r_seq,
                                reference_db_path, working_dir,
                                k=min_kmer_score, l_name=l_name, r_name=r_name)

        if att_data is not None:
            prophage.set_coordinates(att_data[0], att_data[1])
            prophage.set_att_len(len(att_data[3]))
        prophage.detect_orientation()

        prophage.update()
        prophage.clean_record()

        prophage.parent_record.features.append(prophage.feature)
        prophage.parent_record.features.sort(key=lambda x: x.location.start)


def write_prophage_output(outdir, contigs, prophages, tmp_dir, draw):
    """Generates output structure and writes data to file

    :param outdir: Root directory the data will be written to
    :type outdir: pathlib.Path
    :param contigs: Auto-annotated contigs to be written to file
    :type contigs: list
    :param prophages: Identified prophages to be written to file
    :type prophages: list
    :param tmp_dir: where this genome's temporary data are found
    :type tmp_dir: pathlib.Path
    :param draw: draw diagram(s) for this genome's prophage(s)?
    :type draw: bool
    """
    for contig in contigs:
        name = contig.id
        contig.record.annotations = ANNOTATIONS

        genbank_filename = outdir.joinpath(f"{name}.gbk")
        table_filename = outdir.joinpath(f"{name}.csv")

        SeqIO.write(contig.record, genbank_filename, "genbank")

        write_contig_data(contig, table_filename)

    for prophage in prophages:
        name = prophage.id
        prophage_outdir = outdir.joinpath(name)
        prophage_outdir.mkdir(exist_ok=True)

        genbank_filename = prophage_outdir.joinpath(f"{name}.gbk")
        fasta_filename = prophage_outdir.joinpath(f"{name}.fasta")

        SeqIO.write(prophage.record, genbank_filename, "genbank")
        SeqIO.write(prophage.record, fasta_filename, "fasta")

    if draw and prophages:
        draw_complete_diagram(outdir, [contig.record for contig in contigs],
                              prophages, tmp_dir, name=outdir.name)


def write_contig_data(contig, outpath):
    """Generates a csv from data associated with each gene from a bacterial
    sequence contig.

    :param contig: Bacterial sequence contig class
    :type contig: depht.classes.contig.Contig
    :param outpath: Path to the outputted data table file
    :type outpath: pathlib.Path or str
    """
    handle = open(outpath, "w")
    csv_writer = csv.DictWriter(handle, fieldnames=CONTIG_DATA_HEADER)

    csv_writer.writeheader()

    for i, feature in enumerate(contig.genes):
        data = (contig.gene_ids[i], feature.location.start,
                feature.location.end, contig.model_scores[i],
                contig.mask_bits[i], contig.hhsearch_scores[i])

        data_dict = dict()
        for j, header in enumerate(CONTIG_DATA_HEADER):
            data_dict[header] = data[j]

        csv_writer.writerow(data_dict)

    handle.close()


if __name__ == "__main__":
    main()



"""Utility script to retrieve sequences from Genbank using Entrez.
"""
import argparse
import pathlib
import sys

from Bio import SeqIO

from depht_train.functions import entrez

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULTS = {"input_type": "acc", "output_type": "fasta"}

INPUT_TYPES = ["acc", "tax"]
OUTPUT_TYPES = ["fasta", "gb"]


def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("identifiers_file", type=pathlib.Path)
    parser.add_argument("outdir", type=pathlib.Path)

    parser.add_argument("-v", "--verbose", action="store_true")

    parser.add_argument("-it", "--input_type", choices=INPUT_TYPES)
    parser.add_argument("-ot", "--output_type", choices=OUTPUT_TYPES)

    parser.set_defaults(**DEFAULTS)

    return parser.parse_args(unparsed_args)
    

def execute_pull_sequences(identifiers_file, output_dir, input_type="acc",
                           output_type="gb", verbose=False):
    """

    :param identifiers_file:
    :param output_dir:
    :param input_type:
    :param output_type:
    :param verbose:
    :return:
    """
    if not identifiers_file.is_file():
        print(f"File {identifiers_file} could not be found.")
        sys.exit(1)

    # FOR TESTING PLEASE REMOVE IF SEEN IN PRODUCTION
    entrez.set_entrez_credentials(email="chg60@pitt.edu")

    ids = entrez.parse_identifiers_file(identifiers_file)
    if input_type == "tax":

        accessions = set()
        for tax_id in ids:
            if verbose:
                print("Searching for sequence accessions "
                      f"with TaxonID {tax_id} ...")
            
                for acc_id in entrez.esearch_taxa(tax_id):
                    accessions.add(acc_id)
    elif input_type == "acc":
        accessions = ids
    else:
        raise NotImplementedError(f"Input type {input_type} is not supported.")

    output_dir.mkdir(parents=True, exist_ok=True)

    if verbose:
        print("Retrieving sequences with sequence accessions...")
    records = entrez.get_records(list(accessions))

    if verbose:
        print("Writing sequences to file...")
    for record in records:
        output_path = output_dir.joinpath(".".join([record.id, output_type]))

        with output_path.open(mode="w") as filehandle:
            SeqIO.write([record], filehandle, output_type)


def main(unparsed_args=None):
    """Commandline entrypoint to this module"""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    execute_pull_sequences(args.identifiers_file, args.outdir,
                           input_type=args.input_type,
                           output_type=args.output_type,
                           verbose=args.verbose)


if __name__ == "__main__":
    main()



import re

from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------

# Initial HHresult header identification regular expressions
# ==========================================================
HHRESULT_HEADERS = {
                "query_id": re.compile(r"Query         ([\w\W\s\S]+)\n"),
                "match_cols": re.compile(r"Match_columns ([\d]+)\n"),
                "num_seqs": re.compile(r"No_of_seqs    (\d+ out of \d+)\n"),
                "neff": re.compile(r"Neff          ([\d\.]+)\n"),
                "searched_HMMs": re.compile(r"Searched_HMMs ([\d]+)\n"),
                "date": re.compile(r"Date          ([\w\W\s\S]+)\n"),
                "command": re.compile(r"Command       ([\w\W\s\S]+)\n")}

MATCHES_HEADER = (re.compile(
        " No Hit                             Prob E-value P-value  Score    "
        "SS Cols Query HMM  Template HMM\n"))
# ==========================================================

# HHresult data table parsing regular expressions
# ===============================================
# Parses one HHresult hit line start into two capturing groups: No and Hit
TABLE_INDEX = re.compile(
         (r"\s*([\d]+)"  # 'No'/'match_num'
          r"\s*([\.;\S\s\-_\d\w]+)$"))  # 'HitID'/'target_id'

# Parsers one HHresult hit line tail into eleven capturing groups
TABLE_DATA = re.compile(
         (r"\s*(\d+\.\d+)"  # 'Prob'/'probability'
          r"\s*([-+\.Ee\d]+)"  # 'E-value'/'expect'
          r"\s*([-+\.Ee\d]+)"  # 'P-value'/'p_value'
          r"\s*(\d+\.\d+)"  # 'Score'/'score'
          r"\s*([\.\d\-]+)"  # 'SS' secondary structure
          r"\s*(\d+)"  # 'Cols'/'match_cols'
          r"\s*(\d+)-(\d+)"  # 'Query HMM'/('query_start')-('query_end')
          # 'Template HMM'/('hit_start')-('hit_end') ('hit_length)
          r"\s+(\d+)-(\d+)\s*\((\d+)\)\n"))
# ===============================================

# HHresult data body parsing regular expressions
# ==============================================
# Parses one HHresult body hit line start into one capturing group
BODY_INDEX = re.compile(r"No (\d+)")  # 'No'/'match_num'
BODY_NAME = re.compile(r">([\S\s\-_\d\w]+)\n")

# Parses one HHresult body hit line start into eight capturing groups
BODY_DATA = re.compile(
                (r"Probab=([\d\.]+)"  # 'probability'
                 r"\s*E-value=([\d\.Ee\-+]+)"  # 'expect'
                 r"\s*Score=([\d\.]+)"  # 'score'
                 r"\s*Aligned_cols=(\d+)"  # 'match_cols"
                 r"\s*Identities=([\d\.]+)%"  # 'pid'
                 r"\s*Similarity=([-\d\.]+)"  # 'similarity'
                 r"\s*Sum_probs=([\d\.]+|inf)"  # 'sum_probs'
                 r"\s*Template_Neff=([\d\.]+)"))  # 'template_Neff'

ALIGNMENT_BLANK_LINES = 2
# Parses one HHresult body alignment line into four capturing groups
BODY_ALIGNMENT_SEQ = re.compile(r"(\w)\s{1}\S+\s*(\d+)\s{1}([\.\w\-]+)"
                                r"\s*(\d+)\s{1}\(\d+\)")
# Parses one HHresult body alignment line into four capturing groups
BODY_ALIGNMENT_CONS = re.compile(r"(\w)\s{1}Consensus\s*(\d+)\s{1}([\.\w\-~]+)"
                                 r"\s*(\d+)\s{1}\(\d+\)")
# Parses one HHresult body alignment line into three capturing groups
BODY_ALIGNMENT_SS = re.compile(r"(\w)\s{1}([_\w]+)\s*([\-\w])")
# Parses one HHresult body alignment line into one capturing group
BODY_ALIGNMENT_MATCH = re.compile(r"([-+|\.\s]+)")
# Parses one HHresult body alignment line into one capturing group
BODY_ALIGNMENT_CONF = re.compile(r"Confidence\s+([\d\s]+)")

BODY_ALIGNMENT_REGEX_MAP = {"cons": BODY_ALIGNMENT_CONS,
                            "seq": BODY_ALIGNMENT_SEQ,
                            "SS": BODY_ALIGNMENT_SS,
                            "match": BODY_ALIGNMENT_MATCH,
                            "conf": BODY_ALIGNMENT_CONF}
# ==============================================

HHMATCH_TABLE_ATTR = ["match_num", "target_id", "probability", "expect",
                      "p_value", "score", "SS", "match_cols", "query_start",
                      "query_end", "hit_start", "hit_end", "hit_length"]
HHMATCH_BODY_ATTR = ["target_id", "probability", "expect", "score",
                     "match_cols", "pid", "similarity", "sum_probs",
                     "template_Neff"]

HHMATCH_ATTR = list(set(HHMATCH_TABLE_ATTR + HHMATCH_BODY_ATTR))


# ERROR HANDLING
# -----------------------------------------------------------------------------
HHRESULT_BASE_MESSAGE = ("Encountered improper formatting while parsing "
                         "HHResult file.\n")
HHRESULT_ERROR_MESSAGES = {"header": ("HHResult file header "
                                      "could not be recognized"),
                           "table": ("HHResult file data table "
                                     "could not be recognized"),
                           "body": ("HHResult file match result "
                                    "could not be recognized"),
                           "alignment": ("HHResult file match alignment "
                                         "was improperly formatted")}


class InitializationError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class HHResultFormatError(Exception):
    def __init__(self, e, line, line_num):
        line_traceback = "".join([f"at line {line_num}:\n",
                                  f"Line {line_num}> ", line])
        e = " ".join([e, line_traceback])
        super().__init__(e)


# HHRESULT DATA CLASSES
# -----------------------------------------------------------------------------
class HHResult:
    """Class for handling HHsuite result files.
    """
    def __init__(self, filepath):
        self.__filepath__ = filepath
        self.__initialized__ = False
        self.__lcounter = 1

        self.matches = None

        # Sets object attributes in preparatin for HHresult file headers
        for attr_name in HHRESULT_HEADERS.keys():
            setattr(self, attr_name, None)

    def parse_result(self):
        """Parses HHsuite result file.
        """
        filehandle = open(self.__filepath__, "r")

        self.__lcounter = 0
        self._parse_header(filehandle)
        match_index_map = self._parse_table(filehandle)
        self._parse_body(filehandle, match_index_map)

        filehandle.close()

        self.matches = list(match_index_map.values())

    def _parse_header(self, filehandle):
        """Parses HHsuite result file header.

        :param filehandle: An open filehandle for a HHresult-formatted file
        """
        # For each HHresult file header line parse and store information
        for attr_name, reg_expr in HHRESULT_HEADERS.items():
            # Parses HHresult file header line and takes header_group
            header_line = self.__attempt_explicit_read(
                                        filehandle, reg_expr, ltype="header")
            header_split = re.split(reg_expr, header_line)

            # Sets attribute defined from HHRESULT_HEADERS with
            # expression capturing group
            setattr(self, attr_name, header_split[1])

        self.__lcounter += 1
        filehandle.readline()

    def _parse_table(self, filehandle):
        """Parses HHsuite result file table.

        :param filehandle: An open filehandle for a HHresult-formatted file
        :returns: Dictionary mapping HHresult match index to a HHMatch object
        :rtype: dict{}
        """
        self.__attempt_explicit_read(filehandle, MATCHES_HEADER, ltype="table")

        match_index_map = dict()
        # Reads matches until the tail end of a line that matches the
        # TABLE_DATA regular expression cannot be found
        while True:
            table_line, matches = self.__attempt_read_check(
                                            filehandle, TABLE_DATA)
            if not matches:
                break

            match_data_split = re.split(TABLE_DATA, table_line)
            match_id_split = re.split(TABLE_INDEX,
                                      match_data_split.pop(0))

            # Create match and use parsed information while ignoring
            # line end whitespace
            match = HHMatch(self.query_id)
            match.load_from_table_data(match_id_split[1:3] +
                                       match_data_split[0:11])

            match_index_map[match.match_num] = (match)

        return match_index_map

    def _parse_body(self, filehandle, match_index_map):
        """Parses HHsuite result file body.

        :param filehandle: An open filehandle for a HHresult-formatted file
        :param match_index_map:
        :type match_index_map:
                Dictionary mapping HHresult match index to a HHMatch object
        """
        self.__lcounter += 1
        body_line = filehandle.readline()
        while True:
            if re.match(BODY_INDEX, body_line) is None:
                break

            body_index_split = re.split(BODY_INDEX, body_line)

            # Asserts the body line directly after the body index is a
            # name line
            body_line = self.__attempt_explicit_read(
                                    filehandle, BODY_NAME, ltype="body")
            body_name_split = re.split(BODY_NAME, body_line)

            # Asserts the body line directly after the body index is a
            # data line
            body_line = self.__attempt_explicit_read(
                                    filehandle, BODY_DATA, ltype="body")
            body_data_split = re.split(BODY_DATA, body_line)

            hhmatch = match_index_map[body_index_split[1]]
            # Rewrites match data taken from the body of the HHResult file
            # onto the corresponding match, identified by its match index
            hhmatch.load_from_body_data(body_name_split[1:2] +
                                        body_data_split[1:-1])

            # Parses the rest of the body alignment and stores it
            # in a HHAlignment object
            hhalignment, body_line = self._parse_body_alignment(
                                                filehandle, hhmatch.target_id)
            hhmatch.hhalignment = hhalignment

    def _parse_body_alignment(self, filehandle, target_id):
        """Parses HHsuite result file body alignments.

        :param filehandle: An open filehandle for a HHresult-formatted file
        :param target_id: Identifier for the target of the HHR match
        :type target_id: str
        :returns: Returns an HHAlignment object and the last unparsed line
        :rtype hhalignment: pde_utils.hhsuite.HHAlignment
        :rtype aln_line: str
        """
        hhalignment = HHAlignment(self.query_id, target_id)
        blanks = 0
        while True:
            self.__lcounter += 1
            aln_line = filehandle.readline()

            # If there are an equal to or greater amount of blank file lines
            # than specified by ALIGNMENT_BLANK_LINES assume EOF/alignment end
            if aln_line in ["", "\n"]:
                blanks += 1
                if blanks > ALIGNMENT_BLANK_LINES:
                    break
                continue

            # Try to parse a non-blank file line as an alignment line
            # print(self.__lcounter)
            parsed = hhalignment.parse_alignment(aln_line)
            if not parsed:
                break
            else:
                blanks = 0

        hhalignment.compile_alignment()
        return hhalignment, aln_line

    def __attempt_read_check(self, filehandle, regex):
        """Helper function to read in a new line from a file and check
        whether it matches an expected regular expression

        :param filehandle: An open filehandle for a HHresult-formatted file
        :param regex: A regular expression to check a read-line against
        :type regex: re.Pattern
        """
        self.__lcounter += 1
        line = filehandle.readline()

        return (line, (re.search(regex, line) is not None))

    def __attempt_explicit_read(self, filehandle, regex, ltype=None):
        """Helper function to read in a new line from a file and check
        whether it matches an expected regular expression

        :param filehandle: An open filehandle for a HHresult-formatted file
        :param regex: A regular expression to check a read-line against
        :type regex: re.Pattern
        :param ltype: A HHResult file section descriptor
        :type ltype: str
        """
        line = filehandle.readline()
        if re.search(regex, line) is None:
            e = HHRESULT_BASE_MESSAGE + HHRESULT_ERROR_MESSAGES.get(ltype, "")
            raise HHResultFormatError(e, line, self.__lcounter)

        self.__lcounter += 1

        return line

    def check_initialization(self, caller):
        """Safe programming feature - raise an exception if a client
        tries to perform operations on uninitialized object.
        :param caller: name of the method that called this one
        """
        if not self.initialized:
            m = (f"Cannot call method '{caller}' on uninitialized "
                 "HHResult object")

            raise InitializationError(m)


class HHMatch:
    """Class for handling single matches from a HHsuite HHR file.
    """
    def __init__(self, query_id):
        self.query_id = query_id

        self.hhalignment = None

        # Sets object attributes in preparatin for HHresult file matches
        for i in range(len(HHMATCH_ATTR)):
            setattr(self, HHMATCH_ATTR[i], None)

    def load_from_table_data(self, table_data):
        """Loads match data parsed from the table section of a HHResult file.

        :param table_data: Ordered data parsed from the table section of a HHR
        :type table_data: list
        """
        # For each HHresult table match line column parse and store information
        for i in range(len(table_data)):
            setattr(self, HHMATCH_TABLE_ATTR[i], table_data[i])

    def load_from_body_data(self, body_data):
        """Loads match data parsed from the body section of a HHResult file.

        :param table_data: Ordered data parsed from the body section of a HHR
        :type table_data: list
        """
        # For each HHresult body match line column parse and store information
        for i in range(len(body_data)):
            setattr(self, HHMATCH_BODY_ATTR[i], body_data[i])


class HHAlignment:
    """Class for handling alignments for single matches from a HHsuite HHR
    """
    def __init__(self, query_id, target_id):
        self.query_id = query_id
        self.target_id = target_id

        self.query_seq = None
        self._query_record = None
        self.target_seq = None
        self._target_record = None

        self.query_cons_seq = None
        self._query_cons_record = None
        self.target_cons_seq = None
        self._target_cons_record = None

        self.alignment = None

    # Could be improved in the future to handle HHR SS, match qualifiers
    # and match SS per position
    def parse_alignment(self, line):
        """Parses and stores information from a line read from an HHsuite HHR

        :param line: A line from a HHResult file
        :type line: str
        :return:  Returns if given line was recognized as an HHR alignment line
        :rtype: bool
        """
        parsed = False
        for aln_line_type, regex in BODY_ALIGNMENT_REGEX_MAP.items():
            if re.match(regex, line) is not None:
                parsed = True

                aln_split = re.split(regex, line)[1:-1]
                if aln_line_type == "seq":
                    if aln_split[0] == "Q":
                        self._load_seq("query_seq", aln_split[2])
                    elif aln_split[0] == "T":
                        self._load_seq("target_seq", aln_split[2])
                elif aln_line_type == "cons":
                    if aln_split[0] == "Q":
                        self._load_seq("query_cons_seq", aln_split[2])
                    elif aln_split[0] == "T":
                        self._load_seq("target_cons_seq", aln_split[2])
                break

        return parsed

    def _load_seq(self, seq_attr, seq_str):
        """Stores/appends to alignment sequence data from a HHR match

        :param seq_attr: HHAlignment Seq object attribute to load
        :type seq_attr: str
        :param seq_str: Sequence to load/append
        :type seq_str: str
        """
        seq = getattr(self, seq_attr)

        # Initialize Seq object,
        # with gaps if the match start is mid translation
        if seq is None:
            seq_str = seq_str
        else:
            seq_str = str(seq) + seq_str

        setattr(self, seq_attr, Seq(seq_str))

    def compile_alignment(self):
        """Compiles a MultipleSeqAlignment object from a HHR alignment
        """
        if self.query_seq is None or self.target_seq is None:
            return

        records = list()
        self._query_record = SeqRecord(self.query_seq, id=self.query_id)
        records.append(self._query_record)

        if self.query_cons_seq is not None:
            self._query_cons_record = SeqRecord(
                                        self.query_cons_seq,
                                        id=" ".join(
                                            [self.query_id, "Consensus"]))
            records.append(self._query_cons_record)

        if self.target_cons_seq is not None:
            self._target_cons_record = SeqRecord(
                                        self.target_cons_seq,
                                        id=" ".join(
                                            [self.target_id, "Consensus"]))
            records.append(self._target_cons_record)

        self._target_record = SeqRecord(self.target_seq, id=self.target_id)
        records.append(self._target_record)

        self.alignment = MultipleSeqAlignment(records)



"""Utility script to identify conserved phams.
"""
import argparse
import binascii
import pathlib
import sys

import bitarray
from Bio import SeqIO

from depht.data import GLOBAL_VARIABLES
from depht_train.data import PARAMETERS
from depht_train.functions.fileio import (
    read_gene_index_file,
    read_cluster_index_file,
    write_gene_hex_value_file)

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
NAME = GLOBAL_VARIABLES["bacterial_sequences"]["name"]

REP_THRESHOLD = PARAMETERS["shell_db"]["rep_threshold"]

BINARY_TO_HEX_PLACES = 4


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("input_dir", type=pathlib.Path)
    parser.add_argument("output_dir", type=pathlib.Path)
    parser.add_argument("gene_index", type=pathlib.Path)
    parser.add_argument("cluster_index", type=pathlib.Path)

    parser.add_argument("-n", "--name", type=str, default=NAME)
    parser.add_argument("-rt", "--representation_threshold", type=float,
                        default=REP_THRESHOLD)

    return parser.parse_args(unparsed_args)


def screen_conserved_phams(input_dir, output_dir, gene_index, cluster_index,
                           rep_threshold=REP_THRESHOLD, name=NAME):
    """Function to mark putative shell genome content simply by
    identifying subclade-specific conserved gene clusters.

    :param name:
    :param rep_threshold:
    :param input_dir: Path to dir containing protein cluster fasta-alignments
    :type input_dir: pathlib.Path
    :param output_dir:  Path to write conservation hexadecimal output
    :type output_dir: pathlib.Path
    :param gene_index: Path to gene index and relevant metadata file
    :type gene_index: pathlib.Path
    :param cluster_index: Path to a csv table mapping genomes to clusters.
    :type cluster_index: pathlib.Path
    """
    # Read in index and cluster data
    gene_data = read_gene_index_file(gene_index)
    cluster_data = read_cluster_index_file(cluster_index)
    record_cluster_map = get_record_cluster_map(cluster_data)

    # Initialize bitarray list data structure
    gene_rep_bitarrays = [None] * len(gene_data)
    for input_file in input_dir.iterdir():
        if input_file.suffix != ".fasta":
            continue

        gene_records = [record for record in SeqIO.parse(input_file, "fasta")]

        # Ascertain the conservation bitarray value from the members of
        # a protein sequence cluster
        cluster_rep_bitarray = get_cluster_rep_bitarray(
                                            gene_records, gene_data,
                                            cluster_data, record_cluster_map,
                                            rep_threshold)

        for gene_record in gene_records:
            gene_rep_bitarrays[int(gene_record.id)] = cluster_rep_bitarray

    # Convet the bitarrays to hexadecimal representations
    gene_rep_hex_values = [binascii.b2a_hex(bitarray.tobytes())
                           for bitarray in gene_rep_bitarrays]

    # Write the hexadecimal values in order found in the gene index
    gene_hex_value_file = output_dir.joinpath(name).with_suffix(".pbv")
    write_gene_hex_value_file(gene_hex_value_file, gene_rep_hex_values)


def get_cluster_rep_bitarray(gene_records, gene_data, cluster_data,
                             record_cluster_map, rep_threshold):
    """Function to calculate the bitarray value of a protein sequence cluster
    using the frequency of appearence within the clades of their respective
    hosts.

    :param gene_records: SeqRecord objects of the protein sequence cluster
    :type gene_records: list
    :param gene_data: Index and relevant metadata for each protein sequence
    :type gene_data: dict
    :param cluster_data: 2D-array of Sequence IDs grouped by clade membership
    :type cluster_data: list
    :param record_cluster_map: Sequence ID to cluster lookup map.
    :type record_cluster_map: dict
    :param rep_threshold: Cutoff for orthologous proteins to be shell content
    :type rep_threshold: float
    """
    cluster_representation_sets = [set() for _ in range(len(cluster_data))]

    # Create sets of all the host IDs for each protein sequence
    # in the sequence cluster
    for record in gene_records:
        gene_data_dict = gene_data.get(record.id, None)

        if gene_data_dict is None:
            continue

        parent = gene_data_dict["parent"]

        cluster_num = record_cluster_map.get(parent, None)

        if cluster_num is None:
            continue

        cluster_representation_sets[cluster_num].add(parent)

    # Intialize a bitarray to the size of the number of clades
    bit_array_size = len(cluster_data)
    cluster_rep_bitarray = bitarray.bitarray([0] * bit_array_size)

    # Flip bits depending on the frequency of appearence within
    # host sequence clades
    for i, cluster_rep_set in enumerate(cluster_representation_sets):
        representation = len(cluster_rep_set) / len(cluster_data[i])

        if representation < rep_threshold:
            continue

        cluster_rep_bitarray[i] = 1

    return cluster_rep_bitarray


def get_record_cluster_map(cluster_data):
    """

    :param cluster_data:
    :return:
    """
    record_cluster_map = {}

    for index, record_names in enumerate(cluster_data):
        for record_name in record_names:
            record_cluster_map[record_name] = index

    return record_cluster_map


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    screen_conserved_phams(args.input_dir, args.output_dir, args.gene_index,
                           args.cluster_index, name=args.name,
                           rep_threshold=args.representation_threshold)


if __name__ == "__main__":
    main()



"""Standalone script for training just the phage/bacterial classifier
portion of the model."""

import argparse
import pathlib
import pickle
import sys

import pandas as pd
from Bio import SeqIO

from depht.data import GLOBAL_VARIABLES
from depht.functions.annotation import (annotate_record,
                                        cleanup_flatfile_records)
from depht.functions.multiprocess import parallelize, CPUS
from depht.functions.prophage_prediction import build_contig_dataframe
from depht.functions.sniff_format import sniff_format
from depht_train.data import PARAMETERS
from depht_train.functions.train_classifier import train_classifier

DEPHT_DIR = pathlib.Path().home().joinpath(
                            GLOBAL_VARIABLES["model_storage"]["home_dir"])
MODEL_DIR = DEPHT_DIR.joinpath(GLOBAL_VARIABLES["model_storage"]["model_dir"])
WINDOW = PARAMETERS["classifier"]["window"]


def parse_args(unparsed_args):
    """Parse commandline arguments."""
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument("name", type=str,
                   help="one-word name for the resultant model")
    p.add_argument("phages", type=pathlib.Path,
                   help="path to a directory containing phage sequences")
    p.add_argument("bacteria", type=pathlib.Path,
                   help="path to a directory containing bacterial sequences")
    p.add_argument("-p", "--prophage-csv", type=pathlib.Path, default=None,
                   help="path to a 3-column CSV mapping the coordinates of "
                        "any known prophages in the bacterial contigs")
    p.add_argument("-w", "--window-size", type=int, default=WINDOW,
                   help=f"number of genes to average features over "
                        f"[default: {WINDOW}]")
    p.add_argument("-c", "--cpu-cores", type=int, default=CPUS,
                   help=f"number of cpu cores to use [default: {CPUS}]")
    return p.parse_args(unparsed_args)


def get_genome_df(filepath, tmp_dir, window):
    """
    Builds a single bacterial/prophage genome's dataframe from a
    directory of FASTA files.

    :param filepath:
    :type filepath:
    :param tmp_dir:
    :type tmp_dir:
    :param window:
    :type window:
    :return: df
    """
    df_list = list()

    file_fmt = sniff_format(filepath)
    records = sorted([x for x in SeqIO.parse(filepath, file_fmt)],
                     reverse=True, key=lambda x: len(x))

    if file_fmt == "fasta":
        for record in records:
            annotate_record(record, tmp_dir, trna=False)
    else:
        cleanup_flatfile_records(records)

    for record in records:
        if len(record.features) > window:
            df_list.append(build_contig_dataframe(record, window))

    if len(df_list) > 0:
        return pd.concat(df_list, axis=0)


def get_dataframe(filepath, tmp_dir, window, cpus=1):
    """
    Builds a complete bacterial/prophage dataframe from a directory
    of FASTA files.

    :param filepath: path to the FASTAs to integrate into a dataframe
    :type filepath: pathlib.Path
    :param tmp_dir: path where annotation files can be written
    :type tmp_dir: pathlib.Path
    :param window: how many genes to consider in a window
    :type window: int
    :param cpus:
    :type cpus:
    :return: dataframe
    """
    if not tmp_dir.is_dir():
        tmp_dir.mkdir(parents=True)

    jobs = list()
    for f in filepath.iterdir():
        file_fmt = sniff_format(f)

        if file_fmt not in ("fasta", "genbank"):
            continue

        jobs.append((f, tmp_dir, window))

    dataframes = parallelize(jobs, cpus, get_genome_df)

    return pd.concat(dataframes, axis=0)


def train_model(name, phg_dir, bct_dir, window=WINDOW, prophages=None, cpus=1):
    if not phg_dir.is_dir():
        print(f"specified phage directory '{str(phg_dir)}' is not a valid "
              f"directory")
        return

    if not bct_dir.is_dir():
        print(f"specified bacteria directory '{str(bct_dir)}' is not a valid "
              f"directory")
        return

    model_dir = MODEL_DIR.joinpath(name)

    tmp_dir = model_dir.joinpath("tmp")
    tmp_dir.mkdir(exist_ok=True, parents=True)

    phg_csv = model_dir.joinpath(f"_{window}_phage_df.csv")
    if phg_csv.is_file():
        phg_df = pd.read_csv(phg_csv)
    else:
        phg_df = get_dataframe(phg_dir, tmp_dir, window, cpus)
        phg_df["class"] = [1] * len(phg_df)
        phg_df.to_csv(phg_csv, index=False)

    bct_csv = model_dir.joinpath(f"_{window}_bacteria_df.csv")
    if bct_csv.is_file():
        bct_df = pd.read_csv(bct_csv)
    else:
        bct_df = get_dataframe(bct_dir, tmp_dir, window, cpus)
        bct_df["class"] = [0] * len(bct_df)

        if prophages:
            # prevent SettingWithCopyWarning message from appearing
            pd.options.mode.chained_assignment = None

            pro_df = pd.read_csv(prophages, dtype={"contig_id": str,
                                                   "start": int,
                                                   "stop": int})
            for i, prophage in pro_df.iterrows():
                name = prophage["contig_id"]
                start, stop = prophage["start"], prophage["stop"]

                bct_df["class"].loc[
                    (bct_df["contig_id"] == name) &
                    (bct_df["start"] >= start) &
                    (bct_df["stop"] >= start) &
                    (bct_df["start"] <= stop) &
                    (bct_df["stop"] <= stop)] = 1

        bct_df.to_csv(bct_csv, index=False)

    phg_df = phg_df.loc[:, ["ctr_size", "ctr_strand", "class"]]
    bct_df = bct_df.loc[:, ["ctr_size", "ctr_strand", "class"]]

    joint_df = pd.concat([bct_df, phg_df], axis=0)

    clf, figs = train_classifier(joint_df, verbose=True)
    for feature_name, fig in figs:
        filename = model_dir.joinpath(f"{name}_{feature_name}_w{window}.pdf")
        fig.update_layout(title=f"{name}: {feature_name} (w={window})")
        if feature_name == "ctr_size":
            fig.update_xaxes(range=[0, 2000])
        else:
            fig.update_xaxes(range=[0, 35])
        fig.write_image(filename)

    clf_path = model_dir.joinpath("classifier.pkl")
    with open(clf_path, "wb") as classifier_writer:
        pickle.dump(clf, classifier_writer)


def main(unparsed_args=None):
    """Commandline entry point to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    name = args.name
    phg_dir = args.phages
    bct_dir = args.bacteria
    window = args.window_size
    prophages = args.prophage_csv
    cpus = args.cpu_cores

    train_model(name, phg_dir, bct_dir, window=window, prophages=prophages,
                cpus=cpus)


if __name__ == "__main__":
    main()



import random


class KFold:
    """
    Slim implementation of sklearn.model_selection.KFold functionality.
    """
    def __init__(self, n_splits=5, shuffle=False, random_state=None):
        if random_state and not shuffle:
            raise ValueError("random_state cannot be used without shuffle")

        if random_state:
            random.seed(random_state)

        self.n_splits = n_splits
        self.shuffle = shuffle

    def split(self, x):
        """
        Generator that yields the train/test indices for the indicated
        n_splits.

        :param x: the length of the data to be split
        :type x: int
        """
        indices = list(range(x))

        # Shuffle indices if told to do so
        if self.shuffle:
            random.shuffle(indices)

        # Now start creating the splits
        for i in range(self.n_splits):
            train, test = list(), list()
            for x, index in enumerate(indices):
                if x % self.n_splits == i:
                    test.append(index)
                else:
                    train.append(index)
            yield sorted(train), sorted(test)



"""Utility script to plot gene density along a genome.
"""
import argparse
import pathlib
import sys

import matplotlib.pyplot as plt
from Bio import SeqIO

# Defaults for bin width and window size, in kb
BIN_WIDTH = 1000
WINDOW_SIZE = 50000


def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    p = argparse.ArgumentParser(description=__doc__)

    p.add_argument("flatfile_dir", type=pathlib.Path,
                   help="path to a directory of Genbank flatfiles to inspect")

    return p.parse_args(unparsed_args)


def count_genes_per_interval(record, interval=BIN_WIDTH):
    """
    Function to count the number of genes per `interval` nucleotides in the record.

    Counts gene starts, but could easily be adapted to work in some other way

    :param record: a Genbank record to process
    :type record: Bio.SeqRecord.SeqRecord
    :param interval: the number of bases to advance when indexing
    :type interval: int
    :return: genes_per_interval
    """

    genes_per_interval = dict()
    max_interval = len(record) // interval * interval

    # Initialize every interval's counter to 0
    for interval_index in range(0, len(record), interval):
        genes_per_interval[interval_index] = 0

    # Iterate over genes, and place them in their appropriate bin
    for gene in record.features:
        if gene.type == "CDS":
            interval_index = gene.location.start // interval * interval
            genes_per_interval[interval_index] += 1

    # Wrap around ends of the genome
    for interval_index in range(0, WINDOW_SIZE, interval):
        genes_per_interval[interval_index + max_interval] = genes_per_interval[interval_index]
    for interval_index in range(max_interval - WINDOW_SIZE, max_interval, interval):
        genes_per_interval[interval_index - max_interval] = genes_per_interval[interval_index]

    return genes_per_interval


def create_gene_density_plot(interval_counts, record_length, average_density, name, window_size=WINDOW_SIZE, advance=BIN_WIDTH):
    # Now calculate xs and ys for scatterplot
    xs, ys, zs = list(), list(), list()
    for x in range(0, record_length, advance):
        total = 0
        for j in range(x - window_size // 2, x + window_size // 2, advance):
            total += interval_counts[j]
        y = (total / window_size) * advance
        z = (y - average_density) ** 2
        xs.append(x)
        ys.append(y)
        zs.append(z)
    stdev = (float(sum(zs))/len(zs)) ** 0.5
    plt.scatter(xs, ys)
    plt.axhline(y=average_density, color="r", linestyle="-")
    plt.axhline(y=average_density + 2*stdev, color="r", linestyle="--")
    plt.axhline(y=average_density - 2*stdev, color="r", linestyle="--")
    plt.title(f"{name} Gene Density")
    plt.xlabel("Genomic Position (Mbp)")
    plt.ylabel(f"genes/kb over {window_size}kb window")
    plt.ylim(0.25, 2.25)
    return max(ys)


def get_records(flatfile):
    """
    Parses and returns the record from the indicated Genbank flatfile.

    :param flatfile: the path to the flatfile to fetch record(s) from
    :type flatfile: pathlib.Path
    :return: records
    """
    records = list()
    with flatfile.open("r") as fh:
        for record in SeqIO.parse(fh, "genbank"):
            records.append(record)
    return records


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    ff_dir = args.flatfile_dir

    # Make sure Genbank directory exists...
    if not ff_dir.is_dir():
        print(f"input directory '{str(ff_dir)}' does not exist... exiting")
        sys.exit(1)

    # Iterate over ff_dir
    for flatfile in ff_dir.iterdir():
        print(f"Processing {str(flatfile)}...")
        # Parse the Genbank record
        try:
            records = get_records(flatfile)
        except:
            print(f"'{str(flatfile)} is fucked.")
            continue

        if len(records) > 5:
            print(f"'{str(flatfile)}' contains {len(records)} records... is this assembly complete?")

        for record in records:
            if len(record) < WINDOW_SIZE:
                print(f"\tcontig '{record.id}' is too short to inspect for prophages...")
                continue

            avg_dens = float(len([feature for feature in record.features if feature.type == "CDS"])) / len(record) * BIN_WIDTH
            interval_genes = count_genes_per_interval(record)
            max_value = create_gene_density_plot(interval_genes, len(record), avg_dens, flatfile.stem)
            if max_value >= 1.5:
                save_path = flatfile.with_name(f"{flatfile.stem}_x.png")
            else:
                save_path = flatfile.with_suffix(".png")
            plt.savefig(save_path)
            plt.close()


if __name__ == "__main__":
    main()



"""Utility script to recut FASTA sequences to start at a new coordinate.
"""
import argparse
import pathlib
import sys

from depht.functions.fasta import parse_fasta, write_fasta


def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    p = argparse.ArgumentParser(description=__doc__)

    p.add_argument("infile", type=pathlib.Path, help="single-sequence FASTA file to recut")
    p.add_argument("start", type=int, help="which coordinate should become bp 1?")
    p.add_argument("outfile", type=pathlib.Path, help="output filepath")

    return p.parse_args(unparsed_args)


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    infile = args.infile
    new_start = args.start
    outfile = args.outfile

    headers, sequences = parse_fasta(infile)
    if len(headers) != len(sequences):
        raise ValueError(f"{str(infile)} is not a valid FASTA file")
    if len(headers) > 1:
        raise ValueError(f"this script can only be run on single-entry FASTAs")

    sequence = sequences[0]
    sequence = sequence[new_start - 1:] + sequence[:new_start]
    sequences = [sequence]
    write_fasta(headers, sequences, outfile)


if __name__ == "__main__":
    main()



"""Wrapper around a subset of the multiprocessing built-in library
to make running multiple processes easy. Load balancing is left to
the caller."""

import multiprocessing

from depht.classes.progress import show_progress

CPUS = multiprocessing.cpu_count()


def parallelize(inputs, cpus, task, verbose=False):
    """Parallelize some task on an input list across the specified
    number of CPU cores.

    :param inputs: arguments to call `task` with
    :type inputs: list
    :param cpus: number of processor cores to use
    :type cpus: int
    :param task: name of the function to run
    :type task: function
    :param verbose: updating progress bar output?
    :return: results
    """
    results = []

    # Don't do any work if there are no inputs
    if len(inputs) == 0:
        return results

    # User requested some number of cpus - make sure it's sane
    if cpus < 1 or cpus > CPUS:
        cpus = min([CPUS, len(inputs)])

    tasks = []
    for item in inputs:
        if not isinstance(item, tuple):
            item = (item,)
        tasks.append((task, item))

    # Start working on the jobs
    results = start_processes(tasks, cpus, verbose)

    return results


def worker(input_queue, output_queue):
    """Worker function to run jobs from the input queue until a stop
    signal is reached.

    :param input_queue: threadsafe queue to pull jobs from
    :type input_queue: mp.Queue
    :param output_queue: threadsafe queue to add results to
    :type output_queue: mp.Queue
    """
    for func, args in iter(input_queue.get, 'STOP'):
        output_queue.put(func(*args))


def start_processes(inputs, cpus, verbose=False):
    """Spool processes and use them to run jobs.

    :param inputs: jobs to run
    :param cpus: optimized number of processors
    :param verbose: updating progress bar output?
    :return: results
    """
    job_q = multiprocessing.Queue()
    result_q = multiprocessing.Queue()

    # Calculate how often to place a show_progress() job
    frequency = max([1, len(inputs)//100])
    for i, job in enumerate(inputs):
        job_q.put(job)
        if verbose and (i % frequency == 0 or i == len(inputs) - 1):
            job_q.put((show_progress, (i+1, len(inputs))))

    # Set up workers and put a 'STOP' signal at the end of job_q for each
    worker_pool = list()
    for i in range(cpus):
        worker_pool.append(
            multiprocessing.Process(target=worker, args=(job_q, result_q)))
        job_q.put('STOP')

    # Ready... set... go!
    [w.start() for w in worker_pool]

    # Grab results from result_q
    results = []
    for _ in range(len(inputs)):
        result = result_q.get()
        if result is not None:
            results.append(result)

    [worker_n.join() for worker_n in worker_pool]

    return results



import pickle

import pandas as pd

from depht.data import PARAMETERS
from depht.functions.sliding_window import *
from depht.functions.statistics import average

# Number of CDS features to consider in a window
WINDOW = PARAMETERS["prophage_prediction"]["window"]

BACTERIA = 0        # Gene prediction state - bacterial
PROPHAGE = 1        # Gene prediction state - prophage


def average_gene_size(starts, stops, length, window=WINDOW):
    """
    Calculates and returns the average gene size in leading/centered/
    lagging windows of genes.

    :param starts: gene left coordinates
    :type starts: list of int
    :param stops: gene right coordinates
    :type stops: list of int
    :param length: contig length in nucleotides
    :type length: int
    :param window: how many genes to average over
    :type window: int
    :return: leading_sizes, center_sizes, lagging_sizes
    """
    if not len(starts) == len(stops):
        raise ValueError(f"len(starts) ({len(starts)}) != len(stops) ("
                         f"{len(stops)})")

    num_genes = len(starts)

    lagging_sizes = list()
    for i_left, i, i_right in lagging_window(window, num_genes):
        right_coord = stops[i_right]
        left_coord = starts[i_left]

        if right_coord < left_coord:
            right_coord += length

        nucl_dist = right_coord - left_coord + 1
        lagging_sizes.append(round(float(nucl_dist)/window))

    mid_idx = window // 2
    center_sizes = lagging_sizes[mid_idx:]
    center_sizes.extend(lagging_sizes[:mid_idx])

    leading_sizes = lagging_sizes[window:]
    leading_sizes.extend(lagging_sizes[:window])

    return leading_sizes, center_sizes, lagging_sizes


def average_strand_changes(strands, window=WINDOW):
    """
    Calculates and returns the average number of strand changes per
    gene in leading/centered/lagging windows of genes.

    :param strands: gene orientations
    :type strands: list of int
    :param window: how many genes to average over
    :type window: int
    :return: leading_changes, centered_changes, lagging_changes
    """
    num_genes = len(strands)

    lagging_changes = list()
    for i_left, i, i_right in lagging_window(window, num_genes):
        cursor_strand = strands[i_left]
        strand_changes = 0
        for x in range(i_left, i_right + 1):
            if strands[x] != cursor_strand:
                strand_changes += 1
                cursor_strand = strands[x]
        lagging_changes.append(strand_changes)

    mid_idx = window // 2
    center_changes = lagging_changes[mid_idx:]
    center_changes.extend(lagging_changes[:mid_idx])

    leading_changes = lagging_changes[window:]
    leading_changes.extend(lagging_changes[:window])

    return leading_changes, center_changes, lagging_changes


def build_contig_dataframe(contig, window=WINDOW):
    """
    Walks the features of the given contig and calculates the feature
    data to be used for model training and/or model prediction.

    :param contig: a contig to analyze features from
    :type contig: Bio.SeqRecord.SeqRecord
    :param window: the window size to consider genes in
    :type window: int
    :return: pd.DataFrame(feature_dict)
    """
    cds_features = [ftr for ftr in contig.features if ftr.type == "CDS"]

    temp_dict = dict()

    # These are the basis for gene density and strand bias calculations
    lefts, rights, strands = list(), list(), list()
    for feature in cds_features:
        lefts.append(feature.location.start)
        rights.append(feature.location.end)
        strands.append(feature.location.strand)

    lead_len, ctr_len, lag_len = average_gene_size(lefts, rights,
                                                   len(contig), window)
    lead_str, ctr_str, lag_str = average_strand_changes(strands, window)

    temp_dict["contig_id"] = [contig.id] * len(lefts)
    temp_dict["start"] = lefts
    temp_dict["stop"] = rights
    temp_dict["strand"] = strands
    temp_dict["lag_size"] = lag_len
    temp_dict["ctr_size"] = ctr_len
    temp_dict["lead_size"] = lead_len
    temp_dict["lag_strand"] = lag_str
    temp_dict["ctr_strand"] = ctr_str
    temp_dict["lead_strand"] = lead_str

    return pd.DataFrame(temp_dict)


def smooth_by_averaging(values, window_size=25):
    """
    Smooths values by averaging them with their `window_size`
    upstream and downstream neighbors.

    :param values: the values to be smoothed
    :type values: list of float
    :param window_size: number of genes up/down-stream to average over
    :type window_size: int
    :return: smoothed_values
    """
    # Adjust window_size to avoid IndexError on short lists
    window_size = min((window_size, len(values)))
    smoothed_values = list()

    for i in range(len(values)):
        local_values = list()
        min_i, max_i = i - window_size, i + window_size
        for j in range(min_i, max_i + 1):
            if j < 0:
                j += len(values)
            elif j > len(values) - 1:
                j -= len(values)
            local_values.append(values[j])
        smoothed_values.append(average(local_values))

    return smoothed_values


def predict_prophage_genes(contig, classifier, alpha=0.25, min_prob=0.75,
                           mask=None):
    """
    Calculates the gene attributes used by the model to predict
    prophage vs bacterial genes. Then uses the classifier from
    `model_path` to make those predictions.

    :param contig: the contig to make prophage predictions in
    :type contig: depht.classes.contig.Contig
    :param classifier: path to a binary file with prophage classifier inside
    :type classifier: depht.classes.prophage_classifier.ProphageClassifier
    :param alpha: probability above which to keep prophage prediction
    :type alpha: float
    :param min_prob: probability above which prophage signal is considered
    :type min_prob: float
    :param mask: bitwise and will mask known/theorized bacterial genes
    :type mask: list of int
    :return: predictions
    """
    dataframe = build_contig_dataframe(contig.record)

    lead_df = pd.DataFrame()
    lead_df["ctr_size"] = dataframe.loc[:, "lead_size"]
    lead_df["ctr_strand"] = dataframe.loc[:, "lead_strand"]

    center_df = dataframe.loc[:, ["ctr_size", "ctr_strand"]]

    lag_df = pd.DataFrame()
    lag_df["ctr_size"] = dataframe.loc[:, "lag_size"]
    lag_df["ctr_strand"] = dataframe.loc[:, "lag_strand"]

    with open(classifier, "rb") as model_reader:
        classifier = pickle.load(model_reader)

    lead_p = classifier.predict_proba(lead_df)
    center_p = classifier.predict_proba(center_df)
    lag_p = classifier.predict_proba(lag_df)

    predictions = list()
    for x, y, z in zip(lead_p, center_p, lag_p):
        a = float(x + y)/2
        b = float(y + z)/2
        predictions.append(max([a, y, b]))
        # predictions.append(max([x, y, z]))

    # Store model predictions within the contig option
    contig.update_model_scores(predictions)

    predictions = smooth_by_averaging(predictions, window_size=10)

    if mask:
        for gene_i in range(len(predictions)):
            predictions[gene_i] = predictions[gene_i] * mask[gene_i]

    predictions = smooth_by_averaging(predictions, window_size=5)

    prophage_signal = [x >= alpha for x in predictions]

    filter_prophage_signal(prophage_signal, contig.model_scores, min_prob)

    return prophage_signal


def filter_prophage_signal(prophage_signal, predictions, min_prob):
    """Imposes a minimum prophage signal probability on regions in the genome
    with a signal level above alpha.

    :param prophage_signal:
    :type prophage_signal:
    :param predictions:
    :type predictions:
    :param min_prob:
    :type min_prob:
    """
    prophage_block = False
    prophage_block_indicies = list()

    for i in range(len(prophage_signal)):
        prophage_bit_signal = prophage_signal[i]

        if prophage_bit_signal and not prophage_block:
            prophage_block = True
            prophage_block_indicies = [i]

        elif prophage_bit_signal and prophage_block:
            prophage_block_indicies.append(i)

        elif not prophage_bit_signal and prophage_block:
            prophage_block = False

            block_predictions = [predictions[j]
                                 for j in prophage_block_indicies]

            if max(block_predictions) < min_prob:
                for j in prophage_block_indicies:
                    prophage_signal[j] = False

    if prophage_block:
        block_predictions = [predictions[j] for j in prophage_block_indicies]
        if max(block_predictions) < min_prob:
            for j in prophage_block_indicies:
                prophage_signal[j] = False


def predict_prophage_coords(contig, classifier, extend_by=0, mask=None):
    """
    Predicts prophage genes on the contig, then tries to approximate
    the coordinates associated with phage <-> bacterial transitions.

    :param contig: the contig to scan for possible prophages
    :type contig: depht.classes.contig.Contig
    :param extend_by: number of bases to overextend prophages by
    :type extend_by: int
    :param mask: bitwise and will mask known/theorized bacterial genes
    :type mask: list of int
    :return: prophage_coords
    :rtype: list of (int, int)
    """
    gene_predictions = predict_prophage_genes(contig, classifier, mask=mask)

    prophage_coords = list()
    left, right = None, None
    n_inflections = 0
    previous_state = gene_predictions[0]

    for i, current_state in enumerate(gene_predictions):
        if previous_state == BACTERIA and current_state == PROPHAGE:
            left = contig.genes[i].location.start - extend_by
            n_inflections += 1
        elif previous_state == PROPHAGE and current_state == BACTERIA:
            right = contig.genes[i - 1].location.end + extend_by
            prophage_coords.append((left, right))
            left, right = None, None
            n_inflections += 1
        previous_state = current_state
    # Record the last one found
    if left or right:
        prophage_coords.append((left, right))

    # Deal with edge cases (3 major possibilities):
    # Case 0: whole contig is a prophage - last gene processed was a predicted
    # prophage gene, and we had no PROPHAGE <-> BACTERIAL inflections
    if previous_state == PROPHAGE and n_inflections == 0:
        left, right = 0, len(contig.seq)
        prophage_coords.append((left, right))

    # Case 1: there's a prophage that falls off the left edge of a contig -
    # len(prophage_coords) >= 1, and no left coord for first coordinate tuple
    if len(prophage_coords) >= 1 and not prophage_coords[0][0]:
        left, right = 0, prophage_coords[0][-1]
        prophage_coords[0] = (left, right)

    # Case 2: there's a prophage that falls off the right edge of a contig -
    # len(prophage_coords) >= 1, and no right coord for last coordinate tuple
    if len(prophage_coords) >= 1 and not prophage_coords[-1][-1]:
        left, right = prophage_coords[-1][0], len(contig.seq) - 1
        prophage_coords[-1] = (left, right)

    return prophage_coords



"""
Some functions useful for statistics - evaluating models, program
performance, transforming data, etc.
"""
import math


def average(values, kind="arithmetic"):
    """
    Calculate the mean of given values.

    :param values: the values to calculate mean for
    :type values: list of int or list of float
    :param kind: what kind of average to calculate
    :type kind: str
    :return: mean
    """
    if kind == "arithmetic":
        numerator, denominator = sum(values), len(values)
        mean = float(numerator) / denominator
    elif kind == "geometric":
        product, power = math.prod(values), 1.0/len(values)
        mean = product ** power
    elif kind == "harmonic":
        numerator, denominator = len(values), sum([1.0/x for x in values])
        mean = float(numerator) / denominator
    else:
        raise ValueError(f"'{kind}' is not a supported kind of average")

    return mean


def variance(values, mean=None):
    """
    Calculate the variance (sigma squared) of given values.

    :param values: the values to calculate variance for
    :type values: list of int or list of float
    :param mean: the arithmetic average of the given values
    :type mean: int or float
    :return: var
    """
    if not mean:
        mean = average(values)

    var = average([(x - mean) ** 2 for x in values])

    return var


def standard_dev(values, mean=None):
    """
    Calculate the standard deviation (sigma) of given values.

    :param values: the values to calculate variance for
    :type values: list of int or list of float
    :param mean: the arithmetic average of the given values
    :type mean: int or float
    :return: st_dev
    """
    st_dev = variance(values, mean) ** 0.5

    return st_dev


def median(values):
    """
    Calculate the median of given values.

    :param values: the values to calculate variance for
    :type values: list of int or list of float
    :return: med
    """
    temp_values = sorted(values)
    if len(temp_values) % 2 == 0:
        med = average([temp_values[len(temp_values)//2 - 1],
                       temp_values[len(temp_values)//2]])
    else:
        med = temp_values[len(temp_values)//2]

    return med


def true_positive_rate(true_pos, false_neg):
    """
    Calculate the true positive rate.

    NOTE: same as sensitivity, recall, and hit rate

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_neg: number of missed positives
    :type false_neg: int
    :return: tpr
    """
    return float(true_pos)/(true_pos + false_neg)


def false_negative_rate(true_pos, false_neg):
    """
    Calculate the false negative rate.

    NOTE: same as miss rate

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_neg: number of missed positives
    :type false_neg: int
    :return: fnr
    """
    return 1 - true_positive_rate(true_pos, false_neg)


def true_negative_rate(true_neg, false_pos):
    """
    Calculate the true negative rate.

    NOTE: same as specificity, selectivity

    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: tnr
    """
    return float(true_neg)/(true_neg + false_pos)


def false_positive_rate(true_neg, false_pos):
    """
    Calculate the false positive rate.

    NOTE: same as fall-out

    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: fpr
    """
    return 1 - true_negative_rate(true_neg, false_pos)


def positive_predictive_value(true_pos, false_pos):
    """
    Calculate the positive predictive value.

    NOTE: same as precision

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: ppv
    """
    return float(true_pos)/(true_pos + false_pos)


def false_discovery_rate(true_pos, false_pos):
    """
    Calculate the false discovery rate.

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: fdr
    """
    return 1 - positive_predictive_value(true_pos, false_pos)


def negative_predictive_value(true_neg, false_neg):
    """
    Calculate the negative predictive value.

    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :return: npv
    """
    return float(true_neg)/(true_neg + false_neg)


def false_omission_rate(true_neg, false_neg):
    """
    Calculate the false omission rate.

    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :return: for
    """
    return 1 - negative_predictive_value(true_neg, false_neg)


def f1_score(true_pos, false_pos, false_neg):
    """
    Calculate the F1 score.

    NOTE: harmonic mean of TPR and PPV

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :return: f1
    """
    tpr = true_positive_rate(true_pos, false_neg)
    ppv = positive_predictive_value(true_pos, false_pos)
    return 2 * float(tpr * ppv)/(tpr + ppv)


def accuracy(true_pos, false_neg, true_neg, false_pos):
    """
    Calculate the accuracy.

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: acc
    """
    numer = true_pos + true_neg
    denom = true_pos + false_neg + true_neg + false_pos
    return float(numer)/denom


def mcc(true_pos, false_neg, true_neg, false_pos):
    """
    Calculate the Matthews correlation coefficient.

    NOTE: this can be thought of as a balanced accuracy metric
    that isn't skewed by differences in class size.

    :param true_pos: number of correctly identified positives
    :type true_pos: int
    :param false_neg: number of incorrectly identified negatives
    :type false_neg: int
    :param true_neg: number of correctly identified negatives
    :type true_neg: int
    :param false_pos: number of incorrectly identified positives
    :type false_pos: int
    :return: mcc
    """
    numer = (true_pos * true_neg) + (false_pos * false_neg)
    denom = ((true_pos + false_pos) * (true_pos + false_neg) *
             (true_neg + false_pos) * (true_neg + false_neg)) ** 0.5
    return float(numer)/denom


def minmax(values):
    """
    Identify the minimum and maximum values in a list of floats in a
    single pass (as opposed to calling min() and max() separately).

    :param values: a list of floats to find min() and max() values in
    :type values: list of float
    :return: minimum, maximum
    """
    minimum = maximum = None
    for value in values:
        if minimum is None or value < minimum:
            minimum = value
        if maximum is None or value > maximum:
            maximum = value
    return minimum, maximum


def transform(values, min_t=0, max_t=0):
    """
    Transform (in-place) the given values to a new range indicated by
    the range [min_t, max_t], inclusive.

    :param values: the values to transform
    :type values: list
    :param min_t: desired new minimum
    :type min_t: int or float
    :param max_t: desired new maximum
    :type max_t: int or float
    """
    min_o, max_o = minmax(values)

    try:
        for i, value in enumerate(values):
            v = (max_t - min_t) * float(value - min_o)/(max_o - min_o) + min_t
            values[i] = v
    except ZeroDivisionError:
        raise ValueError(f"cannot transform to range [{min_t},{max_t}] because "
                         f"all values are identical")



from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord

from depht.data import GLOBAL_VARIABLES

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
ANNOTATIONS = GLOBAL_VARIABLES["sequences"]["annotations"]
FEATURE_TYPES = GLOBAL_VARIABLES["sequences"]["feature_types"]
DEFAULT_PRODUCT = GLOBAL_VARIABLES["sequences"]["default_product"]

FEATURE_LENGTH_WEIGHT_EXP = 1.5


def realign_subrecord(record, subrecord, subrecord_start, subrecord_end,
                      rev_orient=False):
    """Adds and start-aligns features in the parent record to a subrecord
    given subrecord coordinates with respect to the parent record.

    :param record: Parent record to draw features from.
    :type record: Bio.SeqRecord.SeqRecord
    :param subrecord: Record to place and align features into.
    :type subrecord: Bio.SeqRecord.SeqRecord
    :param subrecord_start: Coordinate start in the record to take features.
    :type subrecord_start: int
    :param subrecord_end: Coordinate end in the record to stop taking features.
    :type subrecord_end: int
    :param rev_orient: Toggles inversion of feature strands in the subrecord.
    :type rev_orient: bool
    """
    for feature in record.features:
        if (feature.location.start < subrecord_start or
                feature.location.end > subrecord_end):
            continue

        if rev_orient:
            feature_start = (subrecord_end - feature.location.end)
            feature_end = (subrecord_end - feature.location.start)
            feature_location = FeatureLocation(feature_start, feature_end)

            if feature.strand == 1:
                strand = -1
            else:
                strand = 1
        else:
            feature_start = (feature.location.start - subrecord_start)
            feature_end = (feature.location.end - subrecord_start)
            feature_location = FeatureLocation(feature_start, feature_end)

            strand = feature.strand

        new_feature = SeqFeature(feature_location, type=feature.type,
                                 qualifiers=feature.qualifiers,
                                 strand=strand)
        subrecord.features.append(new_feature)
        subrecord.features.sort(key=lambda x: x.location.start)


class Prophage:
    """Class to manage Prophage sequences and attributes with respect to
    a parent bacterial sequence.
    """
    def __init__(self, parent_record, seq_id, start=None, end=None, att_len=0,
                 strand=1):
        self.parent_record = parent_record
        self.parent_seq = parent_record.seq
        self.id = seq_id

        self.start = start
        self.end = end
        self.strand = strand

        self.length = 0
        self.feature = None
        self.seq = None
        self.record = None

        self.att_len = att_len
        self.attL = None
        self.attR = None

        self.products = []

        self.update_sequence_attributes()

    def set_coordinates(self, start, end):
        """Sets the coordinates of the prophage in the bacterial sequence

        :param start: The start (left) of the prophage on the sequence.
        :type start: int
        :param end: The end (right) of the prophage on the sequence.
        :type end: int
        """
        self.start = start
        self.end = end

    def set_strand(self, strand):
        """Sets the strand orientation of the prophage in the sequence

        :param strand: The orientation of the prophage in the sequence.
        :type strand: int
        """
        self.strand = strand

        self.update_sequence_attributes()

    def detect_orientation(self):
        """Tries to detect the strand orientation of the prophage in the
        sequence
        """
        if self.record is None:
            return

        orientation = 0
        for feature in self.record.features:
            if feature.type not in FEATURE_TYPES:
                continue

            orientation += (int(feature.strand) *
                            len(feature) ** FEATURE_LENGTH_WEIGHT_EXP)

        if orientation >= 0:
            self.strand = 1
        else:
            self.strand = -1

    def set_att_len(self, att_len):
        """Sets the att length of the prophage in the sequence

        :param att_len: The length of the att sites of the prophage
        :type att_len: int
        """
        self.att_len = att_len

    def update_sequence_attributes(self):
        """Sets prophage SeqFeature and SeqRecord attributes based on
        the object's set coordinates.
        """
        if self.start is None or self.end is None:
            return

        # Sets the prophage SeqFeature according to the stored start and end
        self.feature = SeqFeature(FeatureLocation(self.start, self.end),
                                  strand=self.strand, type="source")
        self.feature.qualifiers["locus_tag"] = [self.id]

        # Extracts the prophage sequence with the created prophage
        self.seq = self.feature.extract(self.parent_seq)
        self.length = len(self.seq)
        # Creates and realigns a Prophage SeqRecord object
        self.record = SeqRecord(self.seq, id=self.id, annotations=ANNOTATIONS)

        realign_subrecord(self.parent_record, self.record,
                          self.start, self.end,
                          rev_orient=(self.strand != 1))

    def update_att_attributes(self):
        """Sets prophage attL and attR attributes based on the object's
        set sequence and att metadata.
        """
        if self.seq is None or self.record is None or self.att_len <= 0:
            return

        left_location = FeatureLocation(0, self.att_len)
        right_location = FeatureLocation(self.length-self.att_len,
                                         self.length)

        if self.strand == 1:
            att_l_location = left_location
            att_r_location = right_location
        else:
            att_l_location = right_location
            att_r_location = left_location

        self.attL = SeqFeature(att_l_location,
                               strand=self.strand, type="misc_recomb")
        self.attL.qualifiers["name"] = [" ".join([self.id, "attL"])]
        self.attL.qualifiers["note"] = [str(self.attL.extract(self.seq))]

        self.attR = SeqFeature(att_r_location,
                               strand=self.strand, type="misc_recomb")
        self.attR.qualifiers["name"] = [" ".join([self.id, "attR"])]
        self.attR.qualifiers["note"] = [str(self.attR.extract(self.seq))]

        self.record.features.append(self.attL)
        self.record.features.append(self.attR)

        self.record.features.sort(key=lambda x: x.location.start)

    def update_products(self):
        """Sets prophage products based on the HMM-HMM profile alignments of
        the translations for the coding features of the prophage.
        """
        if self.seq is None or self.record is None or self.feature is None:
            return

        product_set = set()
        for feature in self.record.features:
            if feature.type != "CDS":
                continue

            product_qualifiers = feature.qualifiers.get("product", None)
            if not product_qualifiers:
                feature.qualifiers["product"] = [DEFAULT_PRODUCT]
                product_qualifiers = feature.qualifiers["product"]

            product = product_qualifiers[0]
            if product == DEFAULT_PRODUCT:
                continue

            product_set.add(product)

        self.products = list(product_set)
        self.feature.qualifiers["note"] = [", ".join(self.products)]

    def update(self):
        """Sets prophage sequence and att features based on stored information
        """
        self.update_products()
        self.update_sequence_attributes()
        self.update_att_attributes()

    def clean_record(self):
        """Add quality-of-life features and tidy prophage extraction artifacts.
        """
        self.record.features.sort(key=lambda x: x.location.start)

        gene_counter = 0
        gene_features = []
        for feature in self.record.features:
            if feature.type not in FEATURE_TYPES:
                continue

            gene_counter += 1
            locus_tag = "_".join([self.id, str(gene_counter)])

            feature.qualifiers["locus_tag"] = [locus_tag]
            feature.qualifiers["gene"] = [gene_counter]

            gene_feature = SeqFeature(feature.location,
                                      type="gene")
            gene_feature.qualifiers["gene"] = [gene_counter]
            gene_feature.qualifiers["locus_tag"] = [locus_tag]
            gene_features.append(gene_feature)

        source_feature = SeqFeature(FeatureLocation(0, self.length),
                                    type="source")

        self.record.features = self.record.features + gene_features
        self.record.features.append(source_feature)

        self.record.features.sort(key=lambda x: x.location.start)
